import socket
import time
from _socket import timeout
from threading import Thread
import re
import numpy as np
import serial

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceGalil(DeviceSocket):
    terminator = b":"
    analog_input_readout_command = (
        "MG @AN[0],@AN[1],@AN[2],@AN[3],@AN[4],@AN[5],@AN[6],@AN[7];"
    )
    analog_output_readout_command = (
        "MG @AO[0],@AO[1],@AO[2],@AO[3],@AO[4],@AO[5],@AO[6],@AO[7];"
    )
    analog_output_limits_command = "MG _DQ0,_DQ1,_DQ2,_DQ3,_DQ4,_DQ5,_DQ6,_DQ7;"
    analog_input_limits_command = "MG _AQ0,_AQ1,_AQ2,_AQ3,_AQ4,_AQ5,_AQ6,_AQ7;"
    digital_lO_readout_command = "MG _TI0,_TI1,_OP0,_OP1;"
    push_commands_max = 1
    response_timeout = 1000
    digital_mask = (1, 2, 4, 8, 16, 24, 48)

    def __init__(
        self, host, port, udp_host=None, udp_port=None, device_id=None, config=None
    ):
        self.udp_host = udp_host
        self.udp_port = udp_port
        self.udp_socket = None
        self.DO_config = []
        self.DI_config = []
        self.AO_config = []
        self.AI_config = []
        self.analog_input_date_type = np.dtype(
            [
                ("AI0", np.ushort),
                ("AI1", np.ushort),
                ("AI2", np.ushort),
                ("AI3", np.ushort),
                ("AI4", np.ushort),
                ("AI5", np.ushort),
                ("AI6", np.ushort),
                ("AI7", np.ushort),
            ]
        )
        self.analog_output_date_type = np.dtype(
            [
                ("AO0", np.ushort),
                ("AO1", np.ushort),
                ("AO2", np.ushort),
                ("AO3", np.ushort),
                ("AO4", np.ushort),
                ("AO5", np.ushort),
                ("AO6", np.ushort),
                ("AO7", np.ushort),
            ]
        )
        self.preset_io_attributes()
        super().__init__(host, port, device_id, config)

    def _init_attributes(self):
        super()._init_attributes()

        self.create_attribute(ATTR_MODEL, default_value="", readonly=True)

        self[ATTR_ANALOG_INPUT] = {}
        self[ATTR_ANALOG_OUTPUT] = {}
        self[ATTR_DIGITAL_INPUT] = {}
        self[ATTR_DIGITAL_OUTPUT] = {}

        for i in range(8):
            self.create_attribute(
                self.AI_config[i],
                ATTR_ANALOG_INPUT,
                unit="V",
                default_type=np.float16,
                decimals=4,
                readonly=True,
            )
            self.create_attribute(
                self.AO_config[i],
                ATTR_ANALOG_OUTPUT,
                unit="V",
                set_function=lambda value, output=i: self.set_analog_output(
                    output, value
                ),
                default_type=np.float16,
                decimals=4,
            )
            self.create_attribute(
                self.DI_config[i], ATTR_DIGITAL_INPUT, default_type=bool, readonly=True
            )
            self.create_attribute(
                self.DI_config[8 + i],
                ATTR_DIGITAL_INPUT,
                default_type=bool,
                readonly=True,
            )
            self.create_attribute(
                self.DO_config[i],
                ATTR_DIGITAL_OUTPUT,
                default_type=bool,
                set_value_when_set_function=False,
                set_function=lambda value, output=i: self.set_digital_output(
                    output, value
                ),
            )
            self.create_attribute(
                self.DO_config[8 + i],
                ATTR_DIGITAL_OUTPUT,
                default_type=bool,
                set_value_when_set_function=False,
                set_function=lambda value, output=8 + i: self.set_digital_output(
                    output, value
                ),
            )

    def preset_io_attributes(self):
        self.DO_config = [
            "DO[0]",
            "DO[1]",
            "DO[2]",
            "DO[3]",
            "DO[4]",
            "DO[5]",
            "DO[6]",
            "DO[7]",
            "DO[8]",
            "DO[9]",
            "DO[10]",
            "DO[11]",
            "DO[12]",
            "DO[13]",
            "DO[14]",
            "DO[15]",
        ]
        self.DI_config = [
            "DI[0]",
            "DI[1]",
            "DI[2]",
            "DI[3]",
            "DI[4]",
            "DI[5]",
            "DI[6]",
            "DI[7]",
            "DI[8]",
            "DI[9]",
            "DI[10]",
            "DI[11]",
            "DI[12]",
            "DI[13]",
            "DI[14]",
            "DI[15]",
        ]
        self.AO_config = [
            "AO[0]",
            "AO[1]",
            "AO[2]",
            "AO[3]",
            "AO[4]",
            "AO[5]",
            "AO[6]",
            "AO[7]",
        ]
        self.AI_config = [
            "AI[0]",
            "AI[1]",
            "AI[2]",
            "AI[3]",
            "AI[4]",
            "AI[5]",
            "AI[6]",
            "AI[7]",
        ]

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        self.command("EO 0;")

        if self.udp_port is not None:
            self.command(
                "IHC={} <{} >1;".format(self.udp_host.replace(".", ","), self.udp_port)
            )  # Set UDP connection to provided host and port
            self.command("CFC;")  # Set interface C as main output for any UDP packet
            self.command(
                "DR200,2;"
            )  # Start recording data at 200ms and send it to second interface (C)
            self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.udp_socket.bind((self.udp_host, self.udp_port))
            self.udp_socket.settimeout(self.response_timeout * 1e-3)
            Thread(target=self._udp_readout).start()

        self.command("\x12\x16;")  # Get galil model and firmware version
        # self.command("TZ;") # Get all inputs and outputs
        # self.command(self.analog_output_limits_command)
        # self.command(self.analog_input_limits_command)
        self.command("ID;")  # Get input / output configuration
        self.command(self.analog_input_readout_command)
        self.command(self.analog_output_readout_command)
        self.command(self.digital_lO_readout_command, callback=_finish_configuration)

    def _udp_readout(self):
        self.request_timestamp = time.time()
        while self.connected:
            try:
                data, _ = self.udp_socket.recvfrom(1024)  # buffer size is 1024 bytes
                self._handle_udp_readout(data)
                self.response_timestamp = time.time()
            except timeout:
                self.request_timestamp = time.time()
                # DevicePoller.check_device_alive(self)
            except OSError:
                # Socket doesn't exist
                pass
        try:
            self.udp_socket.close()
            self.udp_socket = None
        except AttributeError:
            pass  # udp socket was already closed

    def _handle_udp_readout(self, data):
        try:
            model_name, _revision = self.get_value(ATTR_MODEL).split(" ", maxsplit=1)
            if model_name == "RIO47100":
                for index, ao in enumerate(
                    np.frombuffer(data[8:24], dtype=self.analog_output_date_type)[0]
                ):
                    ao_attribute = self[ATTR_ANALOG_OUTPUT][self.AO_config[index]]
                    ao_min, ao_max = ao_attribute.remove_offset_factor(
                        ao_attribute[MIN]
                    ), ao_attribute.remove_offset_factor(ao_attribute[MAX])
                    dt_info = np.iinfo(self.analog_output_date_type[index])
                    self.set_raw_value(
                        (ATTR_ANALOG_OUTPUT, self.AO_config[index]),
                        (
                            ((ao - dt_info.min) * (ao_max - ao_min))
                            / (dt_info.max - dt_info.min)
                        )
                        + ao_min,
                    )

                for index, ai in enumerate(
                    np.frombuffer(data[24:40], dtype=self.analog_input_date_type)[0]
                ):
                    ai_attribute = self[ATTR_ANALOG_INPUT][self.AI_config[index]]
                    ai_min, ai_max = ai_attribute.remove_offset_factor(
                        ai_attribute[MIN]
                    ), ai_attribute.remove_offset_factor(ai_attribute[MAX])
                    dt_info = np.iinfo(self.analog_input_date_type[index])
                    self.set_raw_value(
                        (ATTR_ANALOG_INPUT, self.AI_config[index]),
                        (
                            ((ai - dt_info.min) * (ai_max - ai_min))
                            / (dt_info.max - dt_info.min)
                        )
                        + ai_min,
                    )

                digital_output = format(
                    np.frombuffer(data[40:42], dtype=np.ushort)[0], "b"
                )
                digital_input = format(
                    np.frombuffer(data[42:44], dtype=np.ushort)[0], "b"
                )
                digital_output = [bool(int(x)) for x in reversed(digital_output)] + (
                    [False] * (16 - len(digital_output))
                )
                digital_input = [bool(int(x)) for x in reversed(digital_input)] + (
                    [False] * (16 - len(digital_input))
                )

                for i, do in enumerate(digital_output):
                    self.set_raw_value([ATTR_DIGITAL_OUTPUT, self.DO_config[i]], do)
                for i, di in enumerate(digital_input):
                    self.set_raw_value([ATTR_DIGITAL_INPUT, self.DI_config[i]], di)
        except ValueError:
            pass

    def start_polling(self):
        if self.udp_port is None:
            Device.start_polling(self)
            self.poll_command(self.analog_input_readout_command, 200)
            self.poll_command(self.analog_output_readout_command, 200)
            self.poll_command(self.digital_lO_readout_command, 200)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = super().handle_readout(readout_buffer)
        # Strip white characters from output
        output = output.strip()
        if output and output[0] == "?":
            self.handle_command_error(command, output)
        elif command == "TC 1;":
            readout_buffer = output
            self.handle_command_error(command, output)
        elif command == self.analog_input_readout_command:
            analog_inputs = map(float, output.split())
            for i, value in enumerate(analog_inputs):
                self.set_raw_value((ATTR_ANALOG_INPUT, self.AI_config[i]), value)
        elif command == self.analog_output_readout_command:
            analog_outputs = map(float, output.split())
            for i, value in enumerate(analog_outputs):
                self.set_raw_value((ATTR_ANALOG_OUTPUT, self.AO_config[i]), value)
        elif command == self.digital_lO_readout_command:
            di0, di1, do0, do1 = map(lambda v: int(float(v)), output.split())
            for bit in range(8):
                self.set_raw_value(
                    (ATTR_DIGITAL_INPUT, self.DI_config[bit]), bool(di0 & (1 << bit))
                )
                self.set_raw_value(
                    (ATTR_DIGITAL_INPUT, self.DI_config[8 + bit]),
                    bool(di1 & (1 << bit)),
                )
                self.set_raw_value(
                    (ATTR_DIGITAL_OUTPUT, self.DO_config[bit]), bool(do0 & (1 << bit))
                )
                self.set_raw_value(
                    (ATTR_DIGITAL_OUTPUT, self.DO_config[8 + bit]),
                    bool(do1 & (1 << bit)),
                )
        elif command == "ID;":
            for line in output.split("\r\n"):
                if line.startswith("analog inputs"):
                    if "programmable range" in line:
                        self.command(self.analog_input_limits_command)
                    else:
                        values = re.search(
                            r"analog inputs ([0-9]+)-([0-9]+) = [0-9]+ bits range ([0-9]+) to ([0-9]+)",
                            line,
                        )
                        if values is not None:
                            first_input, last_input, min_value, max_value = (
                                values.group(1),
                                values.group(2),
                                values.group(3),
                                values.group(4),
                            )
                            for index in range(int(first_input), int(last_input) + 1):
                                if self.get_attribute(
                                    (ATTR_ANALOG_INPUT, self.AI_config[index], MIN)
                                ) in (None, np.inf, -np.inf):
                                    self.set_raw_value(
                                        (ATTR_ANALOG_INPUT, self.AI_config[index]),
                                        float(min_value),
                                        key=MIN,
                                    )
                                if self.get_attribute(
                                    (ATTR_ANALOG_INPUT, self.AI_config[index], MAX)
                                ) in (None, np.inf, -np.inf):
                                    self.set_raw_value(
                                        (ATTR_ANALOG_INPUT, self.AI_config[index]),
                                        float(max_value),
                                        key=MAX,
                                    )
                if line.startswith("analog outputs"):
                    if "programmable range" in line:
                        self.command(self.analog_output_limits_command)
                    else:
                        values = re.search(
                            r"analog outputs ([0-9]+)-([0-9]+) = [0-9]+ bits range ([0-9]+) to ([0-9]+)",
                            line,
                        )
                        if values is not None:
                            first_input, last_input, min_value, max_value = (
                                values.group(1),
                                values.group(2),
                                values.group(3),
                                values.group(4),
                            )
                            for index in range(int(first_input), int(last_input) + 1):
                                if self.get_attribute(
                                    (ATTR_ANALOG_OUTPUT, self.AO_config[index], MIN)
                                ) in (None, np.inf, -np.inf):
                                    self.set_raw_value(
                                        (ATTR_ANALOG_OUTPUT, self.AO_config[index]),
                                        float(min_value),
                                        key=MIN,
                                    )
                                if self.get_attribute(
                                    (ATTR_ANALOG_OUTPUT, self.AO_config[index], MAX)
                                ) in (None, np.inf, -np.inf):
                                    self.set_raw_value(
                                        (ATTR_ANALOG_OUTPUT, self.AO_config[index]),
                                        float(max_value),
                                        key=MAX,
                                    )
        elif command == "\x12\x16;":
            self.set_raw_value(ATTR_MODEL, output)
        elif command == self.analog_output_limits_command:
            analog_outputs_limits = output.strip().split()
            # This is analog output channels limits mapping from Galil doc
            analog_limit_ranges = [(-10, 10), (0, 5), (0, 10), (-5, 5), (-10, 10)]
            new_date_type = []
            for index, limit_range_option in enumerate(analog_outputs_limits):
                # Get lower and uper limit for each analog output channel
                lower_limit, upper_limit = analog_limit_ranges[
                    int(float(limit_range_option))
                ]
                # Figure out data type
                dt = np.ushort
                # Append tuple representing analog output data type
                new_date_type.append((self.AO_config[index], dt))
                # Set both limits for analog channel
                if self.get_attribute(
                    (ATTR_ANALOG_OUTPUT, self.AO_config[index], MIN)
                ) in (None, np.inf, -np.inf):
                    self.set_raw_value(
                        (ATTR_ANALOG_OUTPUT, self.AO_config[index]),
                        lower_limit,
                        key=MIN,
                    )
                if self.get_attribute(
                    (ATTR_ANALOG_OUTPUT, self.AO_config[index], MAX)
                ) in (None, np.inf, -np.inf):
                    self.set_raw_value(
                        (ATTR_ANALOG_OUTPUT, self.AO_config[index]),
                        upper_limit,
                        key=MAX,
                    )
            # Set numpy data type, which is used to parse binary UDP readout
            self.analog_output_date_type = np.dtype(new_date_type)
        elif command == self.analog_input_limits_command:
            analog_inputs_limits = output.strip().split()
            # This is analog input channels limits mapping from Galil doc
            analog_limit_ranges = [(-10, 10), (-5, 5), (-10, 10), (0, 5), (0, 10)]
            new_date_type = []
            for index, limit_range_option in enumerate(analog_inputs_limits):
                # Get lower and uper limit for each analog output channel
                lower_limit, upper_limit = analog_limit_ranges[
                    int(float(limit_range_option))
                ]
                # Figure out data type
                dt = np.short if lower_limit < 0 else np.ushort
                # Append tuple representing analog output data type
                new_date_type.append((self.AO_config[index], dt))
                # Set both limits for analog channel
                if self.get_attribute(
                    (ATTR_ANALOG_INPUT, self.AI_config[index], MIN)
                ) in (None, np.inf, -np.inf):
                    self.set_raw_value(
                        (ATTR_ANALOG_INPUT, self.AI_config[index]), lower_limit, key=MIN
                    )
                if self.get_attribute(
                    (ATTR_ANALOG_INPUT, self.AI_config[index], MAX)
                ) in (None, np.inf, -np.inf):
                    self.set_raw_value(
                        (ATTR_ANALOG_INPUT, self.AI_config[index]), upper_limit, key=MAX
                    )
            # Set numpy data type, which is used to parse binary UDP readout
            self.analog_input_date_type = np.dtype(new_date_type)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(
                self.device_id, [command, readout_buffer], token
            )

        return command, output, callback, token

    def set_analog_output(self, output, value, callback=None):
        """
        Set analog output to value 0-10V.

        :param output: int
        :param value: float
        :param callback: callable
        """
        self.logger.debug("Set analogue AO{} to value {}".format(output, value))
        return self.command("AO{},{};".format(output, value), callback, with_token=True)

    def set_digital_output(self, output, value, callback=None):
        """
        Set digital output to value 0 or 1.

        :param output: int
        :param value: bool
        :param callback: callable
        """
        self.logger.debug("Set digital output to value {}".format(value))
        if value == 0:
            return self.command("CB{};".format(output), callback, with_token=True)
        else:
            return self.command("SB{};".format(output), callback, with_token=True)

    @expose_method({"bitmask": "0-65535"})
    def set_digital_output_mask(self, bitmask):
        self.logger.debug("Set digital output mask to value {}".format(bitmask))
        bitmask = int(bitmask)
        self.command("OP{};".format(bitmask))

    def set_command_echo(self, value, callback=None):
        r"""
        Set printing command for serial port together with command output.

        Example::

            EO 0: MG 1; => '1.0000\r\n:'

            EO 1: MG 1; => 'MG1; 1.0000\r\n:'

        :param value:
        :param callback: callable
        :return:
        """
        self.logger.debug("Set command echo to value {}".format(value))
        return self.command("EO {};".format(value), callback, with_token=True)

    def valid_command_format(self, command):
        return command is not None and command[-1] == ";"

    def collect_incoming_data(self, data: bytes) -> None:
        datastr = data.decode("utf-8")
        if datastr.find("?") != -1:
            (
                _attribute,
                token,
                callback,
                _returning,
            ), _command_init_timestamp = self.commands_buffer.popleft()
            self.command("TC 1;", with_token=token, callback=callback)
            datastr = datastr[datastr.find("?") + 2 :]

        self.buffer.append(datastr)

    def get_error_code(self, callback=None):
        self.command("TC 1;", callback=callback)

    def close_connection(self):
        super().close_connection()
        if self.udp_socket is not None:
            self.udp_socket.close()


# abstract methods `collect_incoming_data` and `found_terminator` are defined in
# DeviceSocket (base class of DeviceGalil). Since multiple inheritance is used here
# we better disable the pylint warning.
# pylint: disable=abstract-method
# see https://github.com/python/mypy/issues/9319 for mypy error
class DeviceGalilPort(DevicePort, DeviceGalil):  # type: ignore
    terminator = DeviceGalil.terminator

    # Not sure _why_ super init isn't called though.
    # pylint: disable=super-init-not-called
    def __init__(
        self,
        port,
        baud_rate=115200,
        parity=serial.PARITY_NONE,
        stop_bits=serial.STOPBITS_ONE,
        byte_size=serial.EIGHTBITS,
        device_id=None,
        udp_host=None,
        udp_port=None,
        config=None,
    ):
        self.udp_host = udp_host
        self.udp_port = udp_port
        self.udp_socket = None
        self.DO_config = []
        self.DI_config = []
        self.AO_config = []
        self.AI_config = []
        self.analog_input_date_type = np.dtype(
            [
                ("AI0", np.ushort),
                ("AI1", np.ushort),
                ("AI2", np.ushort),
                ("AI3", np.ushort),
                ("AI4", np.ushort),
                ("AI5", np.ushort),
                ("AI6", np.ushort),
                ("AI7", np.ushort),
            ]
        )
        self.analog_output_date_type = np.dtype(
            [
                ("AO0", np.ushort),
                ("AO1", np.ushort),
                ("AO2", np.ushort),
                ("AO3", np.ushort),
                ("AO4", np.ushort),
                ("AO5", np.ushort),
                ("AO6", np.ushort),
                ("AO7", np.ushort),
            ]
        )
        self.preset_io_attributes()
        DevicePort.__init__(
            self, port, baud_rate, parity, stop_bits, byte_size, device_id, config
        )

    def handshake(self):
        self.logger.info("Handshake initiated")
        self.push(b"MG 1;")
        time.sleep(0.02)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            if self.read_all().strip() in ("1.0000\r\n:", "MG 1; 1.0000\r\n:"):
                return True
        return False



import time
from threading import Thread, Event
from time import sleep

import numpy as np

import kamzik3
from kamzik3 import units, DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetDataAccess import get_from_dict, set_in_dict
from kamzik3.snippets.snippetsDecorators import expose_method

try:
    from karabo_bridge import Client as KbClient
except ImportError:
    raise DeviceError("karabo_bridge import error!")


class DeviceKaraboBridge(Device):
    karabo_client = None

    def __init__(self, host, port, device_id=None, config=None):
        self.host = host
        self.port = port
        self.stop_auto_refresh = Event()
        self.counter_running = False
        self.stop_auto_refresh.clear()
        super().__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(
            ATTR_REFRESHING, default_value=False, readonly=True, default_type=bool
        )

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...

        :param args: connect attributes
        """
        self.connecting = True
        self.karabo_client = KbClient("tcp://{}:{}".format(self.host, self.port))
        self.handle_configuration_event()
        self.connected = True
        self.connecting = False

    def handle_configuration(self):
        start_at = time.time()
        self._update_attributes()
        self._config_commands()
        self._config_attributes()
        self.start_polling()
        self.set_status(STATUS_CONFIGURED)
        self.auto_refresh()
        self.logger.info(
            "Device configuration took {} sec.".format(time.time() - start_at)
        )

    def _update_attributes(self):
        """
        To make things general, I'm just getting one next() call from Karabo bridge.

        Saving for now only floats and integer values.
        Ignoring numpy arrays.
        """
        data, metadata = self.karabo_client.next()

        for key, value in data.items():
            self.attributes[key] = {}
            decimals = 0

            for sub_key, sub_value in value.items():
                if isinstance(sub_value, int):
                    default_type = np.int64
                elif isinstance(sub_value, float):
                    default_type = np.float64
                    decimals = 6
                else:
                    continue
                self.create_attribute(
                    sub_key,
                    key,
                    default_value=sub_value,
                    readonly=True,
                    default_type=default_type,
                    decimals=decimals,
                )

        self.create_attribute(
            ATTR_MIN_UPDATE_TIMEOUT,
            default_value=0,
            default_type=np.float64,
            unit="ms",
            decimals=0,
        )
        self.create_attribute(
            ATTR_TRAIN_TIMEOUT,
            default_value=100,
            default_type=np.float64,
            unit="ms",
            decimals=0,
        )
        self.create_attribute(
            ATTR_REMOTE_TIMESTAMP,
            default_type=np.float64,
            unit="sec",
            readonly=True,
            decimals=6,
        )
        self.create_attribute(
            ATTR_LOCAL_TIMESTAMP,
            default_type=np.float64,
            unit="sec",
            readonly=True,
            decimals=6,
        )
        self.create_attribute(
            ATTR_DELTA_TIMESTAMP,
            default_value=0,
            default_type=np.int64,
            unit="ms",
            readonly=True,
            decimals=6,
        )
        self.create_attribute(
            ATTR_COUNTER,
            default_value=0,
            default_type=np.uint16,
            set_value_when_set_function=False,
            set_function=self.count_trains,
        )
        self.create_attribute(
            ATTR_LOCAL_TRAIN_ID, default_value=0, default_type=np.int64, readonly=True
        )
        self.create_attribute(
            ATTR_REMOTE_TRAIN_ID, default_value=0, default_type=np.int64, readonly=True
        )
        self.create_attribute(ATTR_TRAIN_OFFSET, default_value=0, default_type=np.int64)

        self.update_metadata(list(metadata.values())[0])

    def _set(self, attribute, value, callback=None):
        """
        This sets Device attribute.

        Use this function when You want to set attribute by tuple or list key.
        Example: Device.set((ATTR_STATUS, VALUE), STATUS_IDLE)
        Attribute value is pushed into server if Device is connected on any.
        To reduce amount of pushed attributes we check if value is different from previous one.

        :param attribute: tuple, list, str
        :param value: mixed
        """
        token = 0
        if isinstance(attribute, (tuple, list)):
            current_value = get_from_dict(self.attributes, attribute)
            if isinstance(value, np.ndarray):
                if np.array_equal(value, current_value):
                    return
            elif value == current_value and attribute[0] != ATTR_COUNTER:
                return

            set_in_dict(self.attributes, attribute, value)
            if attribute[-1] == VALUE:
                # Get new value possibly affected by offset and factor
                value = get_from_dict(self.attributes, attribute)
            token = "{}.{}".format(TOKEN_ATTRIBUTE, ".".join(attribute))
        else:
            current_value = self.attributes[attribute]
            if isinstance(value, np.ndarray):
                if np.array_equal(value, current_value):
                    return
            elif value == current_value and attribute[0] != ATTR_COUNTER:
                return

            self.attributes[attribute] = value
            if attribute == VALUE:
                # Get new value possibly affected by offset and factor
                value = self.attributes[attribute]
            token = "{}.{}".format(TOKEN_ATTRIBUTE, attribute)

        if (
            token
            and kamzik3.session.publisher is not None
            and not isinstance(value, np.ndarray)
        ):
            kamzik3.session.publisher.push_message(
                self.device_id, (attribute, value), token
            )

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def count_trains(self, counts):
        self.reset_counter()
        start_train = self.get_value(ATTR_LOCAL_TRAIN_ID)

        def counter(counts):
            self.set_status(STATUS_BUSY)
            counted = 0
            self.counter_running = True
            while counted < counts and self.counter_running:
                counted = self.get_value(ATTR_LOCAL_TRAIN_ID) - start_train
                self.set_raw_value(ATTR_COUNTER, counted)
                sleep(0.02)
            self.set_raw_value(ATTR_COUNTER, counts)
            self.counter_running = False
            self.set_status(STATUS_IDLE)

        Thread(target=counter, args=[counts]).start()

    @expose_method()
    def reset_counter(self):
        self.set_value(ATTR_COUNTER, 0)

    @expose_method()
    def refresh(self):
        """Get one next() call from Karabo bridge."""
        start_time = time.time()
        self.set_attribute([ATTR_REFRESHING, VALUE], True)

        def _thread():
            self.karabo_next()
            end_time = time.time() - start_time
            self.set_value(ATTR_LATENCY, end_time * 1e3)
            self.set_attribute([ATTR_REFRESHING, VALUE], False)

        Thread(target=_thread).start()

    @expose_method()
    def auto_refresh(self):
        """
        Start routine which will automatically call next() until stopped.

        To save CPU and potentially bandwidth set ATTR_MIN_UPDATE_TIMEOUT attribute.
        Take into account, that next() is blocking!
        """
        self.reset_counter()
        if self.get_attribute([ATTR_REFRESHING, VALUE]):
            return

        self.logger.info("Starting auto refresh routine")
        self.set_attribute([ATTR_REFRESHING, VALUE], True)
        self.stop_auto_refresh.clear()

        def _thread():
            while not self.stop_auto_refresh.is_set():
                start_time = time.time()
                self.karabo_next()
                end_time = time.time() - start_time
                sleep_for = (
                    self[ATTR_MIN_UPDATE_TIMEOUT][VALUE] - end_time * 1e3
                ) / 1e3
                if sleep_for > 0:
                    sleep(sleep_for)
                self.set_value(ATTR_LATENCY, end_time * 1e3)

            self.set_attribute([ATTR_REFRESHING, VALUE], False)
            self.logger.info("Auto refresh routine stopped")

        Thread(target=_thread).start()

    def karabo_next(self):
        """
        Get one next() call from Karabo bridge.

        Take into account, that next() is blocking!
        """
        data, metadata = self.karabo_client.next()
        for key, value in data.items():
            for sub_key, sub_value in value.items():
                if isinstance(sub_value, (int, float)):
                    self.set_value([key, sub_key], sub_value)
        self.update_metadata(list(metadata.values())[0])

    def update_metadata(self, metadata_values):
        """
        This routine is important to get and correct train ID from Karabo bridge.

        ATTR_REMOTE_TRAIN_ID: Karabo train ID
        ATTR_REMOTE_TIMESTAMP: Karabo timestamp
        ATTR_LOCAL_TIMESTAMP: Local computer timestamp
        ATTR_DELTA_TIMESTAMP: Delta timestamp
        (ATTR_LOCAL_TIMESTAMP - ATTR_REMOTE_TIMESTAMP)

        ATTR_TRAIN_OFFSET: User configurable train offset

        ATTR_LOCAL_TRAIN_ID: Interpolated train id using delta timestamp.
        train_id_delta = ATTR_DELTA_TIMESTAMP / ATTR_TRAIN_TIMEOUT
        (ATTR_REMOTE_TRAIN_ID + train_id_delta + ATTR_TRAIN_OFFSET)
        """
        remote_timestamp, local_timestamp = (
            np.float64(metadata_values["timestamp"]),
            time.time(),
        )
        time_delta = (local_timestamp - remote_timestamp) * 1e3
        train_id_delta = (
            units.Quantity(time_delta, "ms") / self[ATTR_TRAIN_TIMEOUT].value()
        )
        self.set_attribute(
            (ATTR_REMOTE_TRAIN_ID, VALUE), metadata_values["timestamp.tid"]
        )
        self.set_attribute((ATTR_REMOTE_TIMESTAMP, VALUE), remote_timestamp)
        self.set_attribute((ATTR_LOCAL_TIMESTAMP, VALUE), np.float64(time.time()))
        self.set_attribute((ATTR_DELTA_TIMESTAMP, VALUE), time_delta)
        self.set_attribute(
            (ATTR_LOCAL_TRAIN_ID, VALUE),
            int(
                round(
                    metadata_values["timestamp.tid"]
                    + train_id_delta.m
                    + self[ATTR_TRAIN_OFFSET][VALUE]
                )
            ),
        )

    @expose_method()
    def stop(self):
        """Stop auto refresh routine by setting stop_auto_refresh flag."""
        self.logger.info("Stopping auto refresh routine")
        self.counter_running = False
        self.stop_auto_refresh.set()

    def close(self):
        self.stop()
        return super().close()



import collections

import pint
import pyqtgraph as pg
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QGridLayout

from kamzik3 import units
from kamzik3.constants import *
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.snippets.snippetsYaml import YamlSerializable

pg.setConfigOption("background", "w")
pg.setConfigOption("foreground", "k")


class AttributePlot(pg.PlotWidget, YamlSerializable):
    sig_repaint = pyqtSignal()

    def __init__(
        self,
        attribute,
        buffer_size,
        plot=None,
        parent=None,
        background="#fff",
        **kwargs
    ):
        super().__init__(parent, background, **kwargs)
        self.current_point = 0
        self.attribute = attribute
        self.x_buffer = collections.deque([], maxlen=buffer_size)
        self.y_buffer = collections.deque([], maxlen=buffer_size)
        try:
            self.base_unit = kwargs.get("unit", attribute[UNIT])
        except (TypeError, KeyError):
            self.base_unit = ""

        pen1 = pg.mkPen(color=(0, 0, 255))
        pen2 = pg.mkPen(color=(0, 0, 0))
        self.value_curve = pg.PlotDataItem(
            self.x_buffer,
            self.y_buffer,
            pen=pen1,
            antialias=False,
            symbol="s",
            symbolSize=2,
            symbolPen=pen2,
        )
        if plot is None:
            self.setLabel("left", kwargs.get("left_label", ""), units=self.base_unit)
            if kwargs.get("bottom_label") is not None:
                self.setLabel("bottom", kwargs.get("bottom_label"))
            if kwargs.get("top_label"):
                self.setLabel("top", kwargs.get("top_label"))
            self.showGrid(x=True, y=True, alpha=0.1)
            self.addItem(self.value_curve)
            self.getAxis("left").enableAutoSIPrefix(False)
            self.main_plot = self
        else:
            plot.setLabel("left", kwargs.get("left_label", ""), units=self.base_unit)
            if kwargs.get("bottom_label") is not None:
                self.setLabel("bottom", kwargs.get("bottom_label"))
            if kwargs.get("top_label"):
                self.setLabel("top", kwargs.get("top_label"))
            plot.showGrid(x=True, y=True, alpha=0.1)
            plot.addItem(self.value_curve)
            plot.getAxis("left").enableAutoSIPrefix(False)
            self.main_plot = plot
        self.sig_repaint.connect(self.slot_set_data)

    def yaml_mapping(self):
        mapping = super().yaml_mapping()
        del mapping["parent"]
        return mapping

    @pyqtSlot()
    def slot_set_data(self):
        if self.main_plot.isVisible():
            if self.main_plot.parent() is None or self.main_plot.parent().isVisible():
                self.value_curve.setData(self.x_buffer, self.y_buffer)

    def update(self, current_point=None, current_value=None):
        if current_point is None:
            self.current_point += 1
        else:
            self.current_point = current_point
        if current_value is None:
            try:
                current_value = self.attribute.value().to(self.base_unit).m
            except (AttributeError, pint.errors.DimensionalityError):
                # Value is None, ignore this point
                return
        else:
            assert isinstance(current_value, units.Quantity)
            current_value = current_value.to(self.base_unit).m
        try:
            self.y_buffer.append(current_value)
            self.x_buffer.append(self.current_point)
        except Exception as e:
            print(e, current_point, current_value)
            return
        self.sig_repaint.emit()

    def close(self):
        self.attribute = None
        self.current_point = 0
        self.reset()
        super().close()

    def reset(self):
        self.x_buffer.clear()
        self.y_buffer.clear()
        self.value_curve.setData([], [])


class DeviceAttributePlot(DeviceWidget):
    def __init__(
        self,
        device,
        attribute,
        buffer_size,
        max_update_rate=300,
        plot=None,
        parent=None,
        background="#fff",
        **kwargs
    ):
        self.attribute = attribute
        self.kwargs = kwargs
        self.buffer_size = buffer_size
        self.max_update_rate = max_update_rate
        self.plot = plot
        self.background = background
        DeviceWidget.__init__(self, device, parent=parent)

    # pylint: disable=unused-argument
    def setupUi(self, form):
        layout = QGridLayout()
        self.setLayout(layout)

    @pyqtSlot()
    def slot_handle_configuration(self):
        """
        Handle widget configuration.

        If overloading make sure to set configured flag.
        """
        self.configured = True
        attribute = self.device.get_attribute(self.attribute)
        if self.kwargs.get("top_label") == "auto":
            self.kwargs["top_label"] = self.device.device_id
        if self.kwargs.get("left_label") == "auto":
            self.kwargs["left_label"] = self.attribute
        attribute_plot = AttributePlot(
            attribute,
            self.buffer_size,
            plot=self.plot,
            background=self.background,
            **self.kwargs
        )
        self.layout().addWidget(attribute_plot)
        attribute.attach_callback(
            lambda _: attribute_plot.update(),
            max_update_rate=self.max_update_rate,
            key_filter=VALUE,
        )



import numpy as np

import kamzik3
from kamzik3 import MacroException
from kamzik3 import session
from kamzik3.constants import *
from kamzik3.devices.general.deviceScanner import DeviceScanner
from kamzik3.macro.macro import Macro
from kamzik3.macro.step import StepDeviceAttributeNumerical
from kamzik3.snippets.snippetsDecorators import expose_method


class DummyScanner(DeviceScanner):
    def _init_attributes(self):
        DeviceScanner._init_attributes(self)
        self.create_attribute(
            ATTR_DWELL_TIME,
            default_value=100,
            unit="ms",
            min_value=0,
            default_type=np.int64,
        )

    def _init_new_scan(self, _scanner_input, _scanner_attributes):
        self.logger.info(
            "Initiating new scan number {}".format(
                kamzik3.session.get_value(ATTR_SCAN_COUNT)
            )
        )

    @expose_method()
    def get_scanner_attributes(self):
        return [ATTR_DWELL_TIME]

    @expose_method()
    def get_scanner_macro(self, scanner_input, scanner_attributes, parent_macro=None):
        macro = Macro(device_id="{}_{}".format(self.device_id, scanner_input.common_id))
        distance = abs(scanner_input.start_point - scanner_input.end_point)
        dwell_time = scanner_attributes[ATTR_DWELL_TIME]
        device_id = scanner_input.step_attributes["device_id"]
        # current_velocity = session.get_device(device_id).get_attribute(ATTR_MAXIMUM_VELOCITY).value()
        maximum_velocity = (
            session.get_device(device_id).get_attribute(ATTR_MAXIMUM_VELOCITY).maximum()
        )
        mv = session.get_device(device_id).get_attribute(ATTR_MAXIMUM_VELOCITY)
        scan_line_time = dwell_time * scanner_input.steps_count
        final_velocity = distance / scan_line_time
        final_velocity = final_velocity.to(maximum_velocity.u)
        if final_velocity > maximum_velocity:
            raise MacroException(
                "Calculated velocity {} is higher than maximum velocity {}.".format(
                    final_velocity, maximum_velocity
                )
            )

        scanner_input.set_steps_count(1)
        if not scanner_input.bidirectional:
            velocity_step_0 = StepDeviceAttributeNumerical(
                "SetMaxVelocityStep",
                device_id,
                ATTR_MAXIMUM_VELOCITY,
                maximum_velocity,
                trigger_log=False,
                negative_tolerance=mv.negative_tolerance(),
                positive_tolerance=mv.positive_tolerance(),
                on_warning=scanner_input.step_attributes["on_warning"],
                timeout=scanner_input.step_attributes["timeout"],
            )
            position_step_0 = StepDeviceAttributeNumerical(
                "SetInitPositionStep",
                device_id,
                ATTR_POSITION,
                scanner_input.start_point,
                trigger_log=False,
                negative_tolerance=scanner_input.step_attributes["negative_tolerance"],
                positive_tolerance=scanner_input.step_attributes["positive_tolerance"],
                on_warning=scanner_input.step_attributes["on_warning"],
                timeout=scanner_input.step_attributes["timeout"],
            )
            macro.add(velocity_step_0)
            macro.add(position_step_0)
        velocity_step_1 = StepDeviceAttributeNumerical(
            "FinalVelocityStep",
            device_id,
            ATTR_MAXIMUM_VELOCITY,
            final_velocity,
            trigger_log=False,
            negative_tolerance=mv.negative_tolerance(),
            positive_tolerance=mv.positive_tolerance(),
            on_warning=scanner_input.step_attributes["on_warning"],
            timeout=scanner_input.step_attributes["timeout"],
        )
        macro.add(velocity_step_1)
        macro.add(scanner_input)
        # velocity_step_2 = StepDeviceAttributeNumerical("PreviousVelocityStep", device_id, ATTR_MAXIMUM_VELOCITY,
        #                                              current_velocity,
        #                                              trigger_log=False, negative_tolerance=mv.negative_tolerance(),
        #                                              positive_tolerance=mv.positive_tolerance())
        # macro.add(velocity_step_2)
        return macro



class Observer:
    """Implement this class to implement Kamzik3 Observer."""

    def subject_update(self, key, value, subject):
        """
        Receive an subject update.

        :param mixed key: subject name
        :param mixed value: subject value
        :param Subject subject: Subject
        """
        raise NotImplementedError("Not implemented")

    def attach_to_subject(self, subject):
        """
        Connect Observer to Subject.

        :param Subject subject: Subject
        """
        subject.attach_observer(self)

    def detach_subject(self, subject):
        """
        Connect Observer to Subject.

        :param Subject subject: Subject
        """
        subject.detach_observer(self)



import time

import numpy as np

from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceLight(Device):
    def __init__(self, device_id=None, config=None):
        super().__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        super()._init_attributes()
        self.create_attribute(
            ATTR_INTENSITY,
            readonly=False,
            default_type=np.float16,
            min_value=0,
            max_value=100,
            unit="%",
            decimals=1,
            set_function=self.set_intensity,
            set_value_when_set_function=False,
        )
        self.create_attribute(
            ATTR_SETPOINT,
            default_value=75,
            readonly=False,
            default_type=np.float16,
            min_value=0,
            max_value=100,
            unit="%",
            decimals=1,
        )

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        _finish_configuration()

    def set_intensity(self, value, callback=None):
        raise NotImplementedError()

    @expose_method()
    # pylint: disable=unused-argument
    def turn_on(self, callback=None):
        self.logger.info("Turn On")
        self.set_intensity(self.get_value(ATTR_SETPOINT))

    @expose_method()
    # pylint: disable=unused-argument
    def turn_off(self, callback=None):
        self.logger.info("Turn Off")
        self.set_intensity(0)

    @expose_method()
    def stop(self):
        self.logger.info("Stop device")
        self.turn_off()


class DeviceLightAttribute(DeviceLight):
    def __init__(
        self,
        device,
        attribute,
        output_switch_attribute=None,
        device_id=None,
        config=None,
    ):
        self.device = device
        self.attribute = Attribute.list_attribute(attribute)
        self.output_switch_attribute = output_switch_attribute
        if output_switch_attribute is not None:
            self.output_switch_attribute = Attribute.list_attribute(
                output_switch_attribute
            )
        super().__init__(device_id, config)

    def connect(self, *args):
        super().connect(*args)
        self.device.attach_attribute_callback(
            self.attribute, self.value_changed, key_filter=VALUE
        )
        self.device.attach_attribute_callback(
            ATTR_STATUS, self.set_status, key_filter=VALUE
        )

    def _init_attributes(self):
        DeviceLight._init_attributes(self)
        if self.output_switch_attribute is not None:
            self.add_attribute(
                ATTR_OUTPUT_SWITCH,
                self.device.get_attribute(self.output_switch_attribute),
            )

    def value_changed(self, value):
        if value is None:
            return
        value_min, value_max = (
            self.device.get_attribute(self.attribute + [MIN]),
            self.device.get_attribute(self.attribute + [MAX]),
        )
        intensity = ((value - value_min) * 100) / (value_max - value_min)
        self.set_offsetted_value(ATTR_INTENSITY, intensity)

    def set_intensity(self, value, callback=None):
        self.logger.info("Set light intensity to {} %".format(value))
        if value > 0:
            self.set_value(ATTR_SETPOINT, value)
        value_min, value_max = (
            self.device.get_attribute(self.attribute + [MIN]),
            self.device.get_attribute(self.attribute + [MAX]),
        )
        attribute_value = (value * (value_max - value_min) / 100) + value_min
        self.device.set_attribute(self.attribute + [VALUE], attribute_value)

    @expose_method()
    def stop(self):
        self.device.stop()

    def close(self):
        self.device.detach_attribute_callback(self.attribute, self.value_changed)
        self.device.detach_attribute_callback(ATTR_STATUS, self.set_status)
        return super().close()

    @expose_method()
    def turn_on(self, callback=None):
        self.logger.info("Turn On")
        if self.output_switch_attribute is not None:
            self.device.set_value(self.output_switch_attribute, True)
        self.set_intensity(self.get_value(ATTR_SETPOINT))

    @expose_method()
    def turn_off(self, callback=None):
        self.logger.info("Turn Off")
        if self.output_switch_attribute is not None:
            self.device.set_value(self.output_switch_attribute, False)
        self.set_intensity(self.get_value(ATTR_SETPOINT))



import re
import time
import numpy as np
import serial

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceZled(DeviceSocket):
    """
    Led light used as a illumination for Microscope
    Documentation is here: https://svnsrv.desy.de/desy/cfel-cxi/doc/manuals/zett_optics/ZLED_communication_protocol.pdf
    Example of config file:

    Zled: !Device:kamzik3.devices.deviceZled.DeviceZledSocket
    device_id: Zled
    port: /dev/ttyACM0
    config:
      attributes:
        !!python/tuple [Setpoint, Value]: 75%

    """

    error_pattern = re.compile("^F00[1-3]$")
    terminator = b"\r"
    push_commands_max = 1
    error_statuses = {
        1: "LED Temperature sensor",
        2: "Undefined",
        3: "LED Connection",
        4: "LED overtemperature",
        5: "PCB overtemperature",
        6: "Fan",
        7: "Undefined",
    }

    def handle_configuration(self):
        """
        Send Command '{0}' to get sensor configuration.
        """
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        # Get Model
        self.command("GN\r")
        # Get serial number
        self.command("G#\r")
        # Get output switch state
        self.command("GO\r")
        # Get SW version
        self.command("GV\r", callback=_finish_configuration)

    def _init_attributes(self):
        """
        Attributes;
            ATTR_SOFTWARE_VERSION: Version of sensor's SW
            ATTR_SERIAL_NUMBER: Serial number
            ATTR_MODEL: Device model
            ATTR_OUTPUT_SWITCH: State of output switch
            ATTR_INTENSITY: Measured intensity
            ATTR_SETPOINT: Intensity setpoint
        """
        DeviceSocket._init_attributes(self)
        self.create_attribute(ATTR_SOFTWARE_VERSION, readonly=True)
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_MODEL, readonly=True)
        self.create_attribute(ATTR_OUTPUT_SWITCH, readonly=True, default_type=bool)
        self.create_attribute(
            ATTR_INTENSITY,
            unit="%",
            default_type=np.int,
            min_value=0,
            max_value=100,
            readonly=False,
            set_function=self.set_intensity,
        )
        self.create_attribute(
            ATTR_SETPOINT,
            default_value=75,
            readonly=False,
            default_type=np.float16,
            min_value=0,
            max_value=100,
            unit="%",
            decimals=1,
        )

    def handle_readout(self, readout_buffer):
        """
        Handle readout from Device.
        :param str readout_buffer: raw readout buffer from Device
        """
        command, output, callback, token = DevicePort.handle_readout(
            self, readout_buffer
        )
        is_error_message = self.error_pattern.match(output)
        if is_error_message is not None:
            # Handle error
            self.set_status(STATUS_ERROR)
            self.handle_command_error(command, readout_buffer)
        elif command == "GS\r":
            # Parse and set output Status
            status_code = int("0x{}".format(output[2:]), 0)
            if status_code == 0:
                self.set_status(STATUS_IDLE)
            else:
                self.set_status(STATUS_ERROR)
                self.handle_command_error(command, self.error_statuses[status_code])
        elif command == "GI\r":
            # Parse and set Intensity
            self.set_raw_value(ATTR_INTENSITY, int(output[2:]))
        elif command == "GO\r":
            # Parse and set Output state
            self.set_raw_value(ATTR_OUTPUT_SWITCH, int(output[2:]))
        elif command == "GV\r":
            # Parse and set SW Version
            self.set_raw_value(ATTR_SOFTWARE_VERSION, output[2:])
        elif command == "GN\r":
            # Parse and set Device model
            self.set_raw_value(ATTR_MODEL, output[2:])
        elif command == "G#\r":
            # Parse and set Serial number
            self.set_raw_value(ATTR_SERIAL_NUMBER, output[2:])

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(
                self.device_id, [str(command), str(readout_buffer)], token
            )

    def start_polling(self):
        DeviceSocket.start_polling(self)
        # Poll status
        self.poll_command("GS\r", 1000)
        # Poll intensity
        self.poll_command("GI\r", 1000)
        # Poll output state
        self.poll_command("GO\r", 1000)

    def set_intensity(self, value, callback=None):
        """
        Set instensity setpoint
        :param Union[int, str] value: intensity in percent (0.0 - 100.0)
        :param Callable callback: Optional callback function
        :return int: token
        """
        self.logger.info("Set intensity to {}%".format(value))
        if value > 0:
            self.set_value(ATTR_SETPOINT, value)
        return self.command("SI{:03}\r".format(int(value)), callback, with_token=True)

    @expose_method()
    def turn_on(self, callback=None):
        """
        Turn the light On
        :param Callable callback: Optional callback function
        :return int: token
        """
        self.logger.info("Turn On")
        self.set_intensity(self.get_value(ATTR_SETPOINT))
        return self.command("SO1\r", callback, with_token=True)

    @expose_method()
    def turn_off(self, callback=None):
        """
        Turn the light Off
        :param Callable callback: Optional callback function
        :return int: token
        """
        self.logger.info("Turn Off")
        self.set_intensity(0)
        return self.command("SO0\r", callback, with_token=True)

    @expose_method()
    def stop(self):
        """
        Just turn Off the light
        :return:
        """
        self.logger.info("Stop device")
        self.turn_off()


# abstract methods `collect_incoming_data` and `found_terminator` are defined in
# DeviceSocket (base class of DeviceZled). Since multiple inheritance is used here
# we better disable the pylint warning.
# pylint: disable=abstract-method
# see https://github.com/python/mypy/issues/9319 for mypy error
class DeviceZledSocket(DevicePort, DeviceZled):  # type: ignore
    """
    Extend ETH controller to enable RS232 connection.
    """

    terminator = b"\r"

    # pylint: disable=super-init-not-called
    def __init__(
        self,
        port,
        baud_rate=9600,
        parity=serial.PARITY_NONE,
        stop_bits=serial.STOPBITS_ONE,
        byte_size=serial.EIGHTBITS,
        device_id=None,
        config=None,
    ):
        DevicePort.__init__(
            self, port, baud_rate, parity, stop_bits, byte_size, device_id, config
        )

    def handshake(self) -> bool:
        """
        Confirm connection to Device via RS232.
        :return bool: True if handshake successful
        """
        self.logger.info("Handshake initiated")
        self.push(b"GS\r")
        time.sleep(0.05)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            output = self.read_all().strip()
            is_error_message = self.error_pattern.match(output)
            return is_error_message is None
        return False



import datetime
import os
import pathlib
import time
from threading import Thread
import numpy as np

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.dummy.deviceDummy import DeviceDummy
from kamzik3.devices.dummy.dummyDetector import DummyDetector
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units


class DeviceDummyEiger(DummyDetector):
    """
    Implementation of a dummy Eiger detector.

    TriggerMode 0 corresponds to "EXTE" in Eiger documentation.
    TriggerMode 3 corresponds to "INTS" in Eiger documentation.
    """

    def __init__(
        self, path, file_writer=None, eiger_stream=None, device_id=None, config=None
    ):
        self.path = path
        self.file_writer = file_writer
        self.eiger_stream = eiger_stream
        DummyDetector.__init__(self, device_id=device_id, config=config)

    def _init_attributes(self):
        DummyDetector._init_attributes(self)
        self.create_attribute("NamePattern", readonly=False)
        self.create_attribute("shotDir", default_value="./")
        self.create_attribute("scanDir", default_value="./")
        self.create_attribute("scanFilenamePattern")
        self.create_attribute(
            "Nimages", default_value=1, default_type=np.uint, min_value=1
        )
        self.create_attribute(
            "Ntrigger", default_value=1, default_type=np.uint, min_value=1
        )
        self.create_attribute(
            ATTR_FRAME_COUNT,
            default_value=0,
            readonly=True,
            description="Internal frame counter",
            default_type=np.uint64,
            min_value=0,
        )
        self.create_attribute(
            "TriggerMode",
            default_value=1,
            default_type=np.uint,
            min_value=1,
            max_value=3,
        )
        self.create_attribute(
            "FrameTime",
            default_value=0.001,
            default_type=np.float,
            min_value=0.001,
            max_value=9999,
            decimals=3,
            unit="s",
        )
        self.create_attribute(
            "CountTime",
            default_value=0.001,
            default_type=np.float,
            min_value=0.001,
            max_value=9999,
            decimals=3,
            unit="s",
            set_function=lambda v: self.set_value(ATTR_EXPOSURE_TIME, v),
        )

    def handle_configuration(self):
        start_at = time.time()
        self._config_attributes()
        self.check_interface()
        if self.file_writer is not None:
            self.share_group(self.file_writer, None, "FileWriter")
        if self.eiger_stream is not None:
            self.share_group(self.eiger_stream, None, "EigerStream")
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(
            "Device configuration took {} sec.".format(time.time() - start_at)
        )

    def check_interface(self):
        """
        Check the configuration of the Eiger interface.

        Either the FileWriter or the Streaming interface has to be enabled (not both).

        Mode 0=enabled, 1=Disabled
        """
        if (
            self.file_writer is not None
            and self.file_writer.get_value("Mode") == "0"
            and self.eiger_stream is not None
            and self.eiger_stream.get_value("Mode") == "0"
        ):
            raise DeviceError(
                "Enable either the File Writer or the Streaming Interface, not both"
            )

        if self.file_writer is None and self.eiger_stream is None:
            raise DeviceError("No interface provided for the Eiger!")

    @expose_method({"Name": "FilePrefix", "Exposure": "FrameTime", "Frames": "Nimages"})
    def acquire_frame(self, Name, Exposure, Frames):
        now = datetime.datetime.now()
        Frames = device_units(self, "Nimages", Frames).m
        self.set_attribute(["TriggerMode", VALUE], 3)

        if self.file_writer is not None:
            path = os.path.join(
                self.get_value("shotDir"),
                f"{Name}_{self.device_id}_{now.strftime('%Y-%m-%d_%H:%M:%S')}",
            )
            self.set_attribute(["FileWriter", "Mode", VALUE], 0)
            self.set_attribute(["FileWriter", "NamePattern", VALUE], path)
            self.set_attribute(["FileWriter", "NimagesPerFile", VALUE], int(Frames))
        else:  # use the Stream subsystem
            path = '{"series_name": "' + Name + '"}'
            self.set_attribute(["EigerStream", "Mode", VALUE], 0)
            self.set_attribute(["EigerStream", "HeaderAppendix", VALUE], path)
            self.set_attribute(["EigerStream", "ImageAppendix", VALUE], path)

        Exposure = device_units(self, "FrameTime", Exposure).m
        # According to Jan 10ms is lowest frame time
        frame_time = 10e-3 if Exposure < 10e-3 else Exposure
        self.set_attribute(["FrameTime", VALUE], frame_time)
        self.set_attribute(["CountTime", VALUE], float(Exposure))
        self.set_attribute(["Ntrigger", VALUE], int(1))
        self.set_attribute(["Nimages", VALUE], int(Frames))
        self.Arm()
        self.Trigger()
        self.set_value(ATTR_FRAME_COUNT, self.get_value(ATTR_FRAME_COUNT) + int(Frames))

    @expose_method({"Name": "FilePrefix", "Exposure": "FrameTime", "Frames": "Nimages"})
    def acquire_scan_frame(self, Name, Exposure, Frames):
        now = datetime.datetime.now()
        Exposure = float(device_units(self, "FrameTime", Exposure).m)
        if Exposure != self.get_value("CountTime"):
            frame_time = 10e-3 if Exposure < 10e-3 else Exposure
            self.set_attribute(["FrameTime", VALUE], frame_time)
            self.set_attribute(["CountTime", VALUE], Exposure)
        Frames = int(float(device_units(self, "Nimages", Frames).m))
        if Frames != self.get_value("Nimages"):
            self.set_attribute(["Nimages", VALUE], Frames)

        self.set_attribute(["TriggerMode", VALUE], 3)
        if self.file_writer is not None:
            path = os.path.join(
                self.get_value("scanDir"),
                f"{Name}_{self.device_id}_{now.strftime('%Y-%m-%d_%H:%M:%S')}",
            )
            self.set_attribute(["FileWriter", "Mode", VALUE], 0)
            self.set_attribute(["FileWriter", "NimagesPerFile", VALUE], Frames)
            self.set_attribute(["FileWriter", "NamePattern", VALUE], path)

        else:  # use the Stream subsystem
            path = '{"series_name": "' + Name + '"}'
            self.set_attribute(["EigerStream", "Mode", VALUE], 0)
            self.set_attribute(["EigerStream", "HeaderAppendix", VALUE], path)
            self.set_attribute(["EigerStream", "ImageAppendix", VALUE], path)

        self.Arm()
        self.Trigger()
        self.set_value(ATTR_FRAME_COUNT, self.get_value(ATTR_FRAME_COUNT) + int(Frames))

    @expose_method(
        {
            "Name": ATTR_FRAME_NAME,
            "Frames": ATTR_SUMMED_FRAMES,
            "Exposure": ATTR_EXPOSURE_TIME,
        }
    )
    def start_acquisition(self, Name, Frames, Exposure):
        self.acquisition_running.set()
        Frames = device_units(self, ATTR_SUMMED_FRAMES, Frames)
        Exposure = device_units(self, ATTR_EXPOSURE_TIME, Exposure)

        self.set_status(STATUS_BUSY)
        self.set_value(ATTR_FRAME_NAME, Name)
        self.set_value(ATTR_SUMMED_FRAMES, Frames)
        self.set_value(ATTR_EXPOSURE_TIME, Exposure)
        self.set_value(ATTR_ACQUIRED_FRAMES, 0)

        def _acquire():
            exposure_time_increment = 10
            current_exposure_time = 0
            last_frame_number = 0
            while self.acquisition_running.isSet():
                time.sleep(exposure_time_increment * 1e-3)
                current_exposure_time += exposure_time_increment
                acquired_frames = int(current_exposure_time / Exposure.m)
                if last_frame_number != acquired_frames:
                    self.set_value(["Output", "Frame"], self._generate_frame())
                    last_frame_number = acquired_frames

                if acquired_frames >= Frames:
                    self.set_value(ATTR_ACQUIRED_FRAMES, Frames)
                    break
                else:
                    self.set_value(ATTR_ACQUIRED_FRAMES, acquired_frames)

            if self.connected:
                self.acquisition_running.clear()
                self.set_status(STATUS_IDLE)

        Thread(target=_acquire).start()

    @expose_method()
    def Trigger(self):
        self.file_writer._create_dir()
        Name = self.get_value(self.get_value(["FileWriter", "NamePattern"]))
        Frames = self.get_value("Nimages")
        Exposure = self.get_value("CountTime") * 1e3
        self.start_acquisition(Name, Frames, Exposure)

    @expose_method()
    def Arm(self):
        self.set_status(STATUS_BUSY)

    @expose_method()
    def Disarm(self):
        self.set_status(STATUS_IDLE)

    @expose_method()
    def Abort(self):
        self.set_status(STATUS_IDLE)

    @expose_method()
    def stop(self):
        if self.get_value(ATTR_STATUS) == STATUS_BUSY:
            self.Abort()
        DummyDetector.stop(self)


class DeviceDummyEigerWriter(DeviceDummy):
    def __init__(self, path, device_id=None, config=None):
        self.path = path
        DeviceDummy.__init__(self, device_id=device_id, config=config)

    def _init_attributes(self):
        DeviceDummy._init_attributes(self)
        self.create_attribute("NimagesPerFile", default_type=np.uint, min=0)
        self.create_attribute("NamePattern")
        self.create_attribute("Mode", default=0)

    def _create_dir(self):
        pathlib.Path(self.get_value("NamePattern")).parent.mkdir(
            parents=True, exist_ok=True
        )



from collections import deque
from time import sleep

import zmq

from kamzik3.constants import WORKER_READY, MSG_TERMINATE, MSG_EMPTY


class Worker:
    """
    This is General Worker Class which can be used as a Server Worker.
    """

    running = False  # Worker running state

    def __init__(self, id, master_host, master_port, zmq_context=None):
        """
        Initiate new Worker.
        Each Worker connects to Master server.
        Worker can perform only instructions defined in instructions_set list.
        :param str id: Worker ID
        :param str master_host:  Host of master server
        :param int master_port: Port of master server
        :param Context zmq_context: ZMQ context
        """
        self.instructions_set = []
        self.zmq_context = zmq_context
        self.master_host = master_host
        self.master_port = master_port
        self.id = id.encode()

    def _init_socket(self):
        """
        Initiate connection to the master server.
        """
        if self.zmq_context is None:
            self.zmq_context = zmq.Context.instance()

        self.socket_worker = self.zmq_context.socket(zmq.REQ)
        self.socket_worker.setsockopt(zmq.IDENTITY, self.id)
        self.socket_worker.connect(
            "tcp://{}:{}".format(self.master_host, self.master_port)
        )

    def run(self):
        """
        Start workers main loop.
        """
        self.running = True
        self._init_socket()
        self.socket_worker.send_multipart([WORKER_READY] + self.instructions_set)
        while self.running:
            request = self.socket_worker.recv_multipart()
            if request[2] == MSG_TERMINATE:
                self.stop()
                continue
            self.socket_worker.send_multipart(
                request[:2] + self.get_payload(request[2:])
            )
        self.socket_worker.close()

    def get_payload(self, request):
        """
        This method gets input data and perform instruction which must be part of the request.
        This function must be implemented in every Worker.
        :param list request: input data
        :return:
        """
        raise NotImplementedError()

    def stop(self):
        """
        Stop Worker.
        Set running state to False.
        """
        self.running = False


class BalancedServer:
    """
    This is main server, which can perform multiple different tasks.
    It's main function is to serve reply to every request of client.
    Every request is passed to the free worker.
    Worker process request, generate reply and server is returning reply to the client.
    In current implementation only one type of worker is supported.
    So one server can serve onl;y one purpose.
    For example acting as a Device Server or File Server.
    Functionality is determined by Worker defined by worker_class.
    """

    running = False  # Server running state

    def __init__(self, host, port, workers_count=5, worker_class=Worker):
        """
        Initiate new Server.
        Functionality is determined by workers.
        :param str host: host
        :param int port: port
        :param int workers_count: Number of workers spawn by server
        :param Class worker_class: Class of Worker
        """
        self.zmq_context = zmq.Context.instance()
        self.host = host
        self.port = port
        self.socket_frontend = None
        self.socket_backend = None
        self.port_backend = None
        self.workers_count = workers_count
        self.worker_class = worker_class
        self.poller = zmq.Poller()
        self.registered_workers = {}
        self.local_workers = []
        self.idle_workers = []

    def _init_frontend(self):
        """
        Initiate frontend part of the server.
        This is visible to the external clients.
        """
        self.socket_frontend = self.zmq_context.socket(zmq.ROUTER)
        self.socket_frontend.setsockopt(
            zmq.IDENTITY, "FE@{}:{}".format(self.host, self.port).encode()
        )
        self.socket_frontend.bind("tcp://{}:{}".format(self.host, self.port))

    def _init_backend(self):
        """
        Initiate backend part of the server.
        This is visible to the workers only.
        """
        self.socket_backend = self.zmq_context.socket(zmq.ROUTER)
        self.socket_backend.setsockopt(
            zmq.IDENTITY, "BE@{}:{}".format(self.host, self.port).encode()
        )
        self.port_backend = self.socket_backend.bind_to_random_port(
            "tcp://{}".format(self.host)
        )
        self.poller.register(self.socket_backend, zmq.POLLIN)

    def _init_workers(self):
        """
        Spawn workers connected to the server backend.
        """
        for i in range(self.workers_count):
            worker = self.worker_class(
                "W_{}".format(i), self.host, self.port_backend, self.zmq_context
            )
            self.local_workers.append(worker)
            worker.start()

    def run(self):
        """
        Start main server loop.
        """
        self.running = True

        # Initiate Frontend, Backend and workers
        self._init_frontend()
        self._init_backend()
        self._init_workers()

        # Start the server
        self.main_loop()

        # Close Frontend and Backend
        self.socket_backend.close()
        self.socket_frontend.close()

    def main_loop(self):
        """
        Main server loop.
        Run until running state is True.
        Poll for Backend and Frontend sockets with any message to read.
        """
        while self.running:
            sockets = dict(self.poller.poll(timeout=10))

            if self.socket_backend in sockets:
                # Handle worker activity on the backend
                request = self.socket_backend.recv_multipart()
                worker, _, client = request[:3]
                if not self.idle_workers:
                    # Poll for clients now that a worker is available
                    self.poller.register(self.socket_frontend, zmq.POLLIN)
                self.idle_workers.append(worker)

                if client != WORKER_READY:
                    # If client reply, send rest back to frontend
                    self.socket_frontend.send_multipart([client] + request[3:])
                else:
                    self.registered_workers[worker] = request[3:]

            if self.socket_frontend in sockets:
                # Get next client request, route to last-used worker
                request = self.socket_frontend.recv_multipart()

                worker = self.idle_workers.pop(0)
                instruction = request[2]
                self.socket_backend.send_multipart([worker, MSG_EMPTY] + request)
                if not self.idle_workers:
                    # Don't poll clients if no workers are available
                    self.poller.unregister(self.socket_frontend)

    def stop(self):
        """
        Stop server main loop.
        Also Terminate every worker on the server.
        """
        for worker in self.local_workers:
            # Send Terminate message to every worker
            self.socket_backend.send_multipart(
                [worker.id, MSG_EMPTY, MSG_EMPTY, MSG_EMPTY, MSG_TERMINATE]
            )

        self.local_workers = []
        self.running = False


class Publisher:
    """
    Simple Publisher loop.
    It's purpose is to collect all messages in the queue and then push it to the socket.
    This is general class for any Publisher.
    _publish_message must be reimplemented.
    """

    running = False  # Publisher running state

    def __init__(self, host, port, zmq_context=None):
        """
        Initiate Publisher.
        :param str host: host
        :param int port: port
        :param Context zmq_context: ZMQ Context
        """
        self.zmq_context = zmq_context
        if self.zmq_context is None:
            self.zmq_context = zmq.Context.instance()
        self.host = host
        self.port = port
        self.messages_queue = deque()
        self._init_socket()

    def _init_socket(self):
        """
        Initiate Publisher socket.
        Use zmq.PUB as we are using PUB -> SUB socket pair.
        """
        self.socket_publisher = zmq.Context.instance().socket(zmq.PUB)
        self.socket_publisher.setsockopt(zmq.IDENTITY, b"P_0")
        self.socket_publisher.bind("tcp://{}:{}".format(self.host, self.port))

    def _publish_message(self, header, message):
        """
        This method is called to publish the message.
        It must be reimplemented in final class.
        :param str header: Header of the message
        :param mixed message: Message depends on the communication protocol
        """
        raise NotImplementedError()

    def run(self):
        """
        Main Publisher loop.
        Until running state is True, publish all messages from the messages_queue.
        :return:
        """
        timeout = 0.005
        self.running = True
        while self.running:
            while self.messages_queue:
                self._publish_message(*self.messages_queue.popleft())
            sleep(timeout)

    def push_message(self, header, message, token=None):
        """
        Push new message into the message queue
        :param str header: Header of the message
        :param mixed message: Message depends on the communication protocol
        :param int token: message token
        """
        try:
            if token is not None:
                header += "." + str(token)
            self.messages_queue.append((header, message))
        except TypeError:
            self.logger.exception("Error pushing message to client")

    def stop(self):
        """
        Stop Publisher loop.
        """
        self.running = False



def device_id_generator():
    """
    Generate ID for unnamed device.

    :return: string
    """
    i = 0
    while True:
        yield "Device_{}".format(i)
        i += 1


def token_generator():
    """
    Generate Token number up to 2**8.

    This token is used for client server communication.

    :return: int
    """
    token, max_token = 1, pow(2, 8)
    while True:
        yield token
        token += 1
        if token >= max_token:
            token = 1


def axis_name_generator(range_of_axis=1):
    """
    Axis name generator for n axis scan.

    :param range_of_axis:
    :type range_of_axis: int
    """
    current = 1
    names = ["Slow", "Slower", "Slowest", "Fast", "Faster", "Fastest"]
    mapping = [
        [3],
        [0, 3],
        [1, 0, 3],
        [1, 0, 3, 4],
        [2, 1, 0, 3, 4],
        [2, 1, 0, 3, 4, 5],
    ]
    col = range_of_axis - 1

    while current <= range_of_axis:
        if range_of_axis <= 6:
            yield "{}".format(names[mapping[col][current - 1]])
        elif current == range_of_axis:
            yield "Fast {}".format(current)
        elif current == 1:
            yield "Slow {}".format(current)
        else:
            yield "{}".format(current)
        current += 1



from kamzik3.devices.observer import Observer


class Subject:
    """Implements Subject which server Observer updates and notifications."""

    _observers = None

    def __init__(self):
        if not self._observers:
            self._observers = []

    def notify(self, key, value):
        """
        Notify all observers with an subject_update.

        :param mixed key: subject name
        :param mixed value: subject value
        """
        # with self.notify_lock:
        for observer in self._observers[:]:
            observer.subject_update(key, value, self)

    def attach_observer(self, observer):
        """
        Attach new observer.

        :param Observer observer: Observer
        """
        if observer not in self._observers:
            if isinstance(observer, Observer):
                self._observers.append(observer)
            else:
                raise AttributeError(
                    "observer has to implement kamzik3.device.observer.Observer interface"
                )

        self.handle_observer_attached(observer)

    def detach_observer(self, observer):
        """
        Detach Observer from list of observers.

        :param Observer observer: Observer
        """
        if observer in self._observers:
            self._observers.remove(observer)
            self.handle_observer_detached(observer)

    def handle_observer_attached(self, observer):
        """
        Hook to be fired after attached Observer.

        :param Observer observer: Observer
        """
        raise NotImplementedError("Has to be implemented")

    def handle_observer_detached(self, observer):
        """
        Hook to be fired after detached Observer.

        :param Observer observer: Observer
        """
        raise NotImplementedError("Has to be implemented")



from __future__ import annotations
import logging
import os
import shutil
import time
from collections import OrderedDict
from threading import RLock
from multiprocessing import Process
from multiprocessing.synchronize import Event as EventType
import numpy as np
import oyaml as yaml
import psutil
from typing import List, Tuple, Union

import kamzik3
from kamzik3 import DeviceUnknownError, DeviceError
from kamzik3.constants import *
from kamzik3.devices.attributeLogger import AttributeLogger, ATTR_LOGGING
from kamzik3.devices.device import Device
from kamzik3.devices.deviceClient import DeviceClient
from kamzik3.devices.general.measurementGroup import MeasurementGroup
from kamzik3.snippets.snippetLogging import (
    set_rotating_file_handler,
    set_sys_exception_handler,
    get_console_handler,
    set_rocket_chat_handler,
)
from kamzik3.snippets.snippetsControlLoops import (
    control_asyncore_loop,
    control_device_poller,
    control_port_read_loop,
    control_device_connection_poller,
)
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsTimer import CallbackTimer
from kamzik3.snippets.snippetsYaml import YamlDumper, YamlConfigSerializer


class DeviceSession(Device):
    """
    Session is mandatory object, that have to exist for any Server or Client.
    It holds reference to any Device or polling loop.
    It must be first to initialize and last to close.
    Session is global and accessible from kamzik3.session object.
    """

    # Log files locations
    LOG_DEVICE_OUTPUT_FILE = "devices.log"
    LOG_CONSOLE_OUTPUT_FILE = "console.log"
    LOG_ATTRIBUTES_OUTPUT_FILE = "attributes.log"
    LOG_GUI_OUTPUT_FILE = "gui.log"
    LOG_MACRO_OUTPUT_FILE = "macro.log"
    # Default log level
    log_level = logging.INFO
    # Lock protecting config file writes
    config_lock = RLock()

    def __init__(self, device_id=None, config=None):
        kamzik3.session = self
        self.devices: OrderedDict[str, Device] = OrderedDict()
        if config is None:
            self.server = None
            self.publisher = None
        else:
            self.server = config.get("server")
            self.publisher = config.get("publisher")
        self.measurement_groups = {}
        # Polling timer of 1s collecting PC resources
        self.polling_timer = CallbackTimer(1000, self.measure_cpu_ram)
        # Current PID of running session
        self.os_process = psutil.Process(os.getpid())
        # Metadata for config file: (Config file location, config dict)
        self.session_config = None
        super().__init__(device_id, config)
        self.connect()

        if self.server is not None:
            self.server.start()
        if self.publisher is not None:
            self.publisher.start()

    def _init_attributes(self):
        """
        Attributes;
            ATTR_LOG_DIRECTORY: Main log directory
            ATTR_RESOURCE_DIRECTORY: Directory for all resources
            ATTR_CACHE_DIRECTORY: Directory for temporary cache
            ATTR_DEVICES_COUNT: Total number of Devices in Session
            ATTR_ALLOW_CONSOLE_LOG: Switch enabling console logging
            ATTR_ALLOW_GUI_LOG: Switch enabling gui logging
            ATTR_ALLOW_DEVICE_LOG: Switch enabling device logging
            ATTR_ALLOW_ATTRIBUTE_LOG: Switch enabling attribute logging
            ATTR_ALLOW_MACRO_LOG: Switch enabling macro logging
            GROUP_RC / ATTR_ALLOW_RC: Switch enabling use of Rocketchat (RC)
            GROUP_RC / ATTR_MAIN_ROOM: Main roop to use for RC
            GROUP_RC / ATTR_AUTH_TOKEN: Authentication token for RC
            GROUP_RC / ATTR_USER_NAME: RC user name
            GROUP_RC / ATTR_PASSWORD: RC password
            GROUP_RC / ATTR_RC_MACRO_DISCUSSION: RC Discussion ID for Macros
            GROUP_RC / ATTR_HOST: RC Host
            ATTR_LOOPS_RUNNING: Flag indicating loops state
            ATTR_CPU_USAGE: Current CPU usage
            ATTR_MEMORY_USAGE: Current Memory usage
            ATTR_AVAILABLE_MEMORY: Total Available Memory
        """
        super()._init_attributes()
        self.create_attribute(
            ATTR_LOG_DIRECTORY, set_function=self.set_log_directory, readonly=True
        )
        self.create_attribute(
            ATTR_RESOURCE_DIRECTORY,
            set_function=self.set_resource_directory,
            readonly=True,
        )
        self.create_attribute(
            ATTR_CACHE_DIRECTORY, set_function=self.set_cache_directory, readonly=True
        )
        self.create_attribute(
            ATTR_DEVICES_COUNT, default_value=0, default_type=np.uint64, readonly=True
        )
        self.create_attribute(
            ATTR_ALLOW_CONSOLE_LOG, default_value=True, default_type=bool
        )
        self.create_attribute(ATTR_ALLOW_GUI_LOG, default_value=True, default_type=bool)
        self.create_attribute(
            ATTR_ALLOW_DEVICE_LOG, default_value=True, default_type=bool
        )
        self.create_attribute(
            ATTR_ALLOW_ATTRIBUTE_LOG, default_value=True, default_type=bool
        )
        self.create_attribute(
            ATTR_ALLOW_MACRO_LOG, default_value=True, default_type=bool
        )
        self.create_attribute(
            ATTR_ALLOW_RC,
            group=GROUP_RC,
            default_value=False,
            readonly=True,
            default_type=bool,
        )
        self.create_attribute(
            ATTR_MAIN_ROOM,
            group=GROUP_RC,
            default_value=self.device_id.lower(),
            readonly=True,
        )
        self.create_attribute(
            ATTR_AUTH_TOKEN,
            group=GROUP_RC,
            default_value=None,
            readonly=True,
            display=False,
        )
        self.create_attribute(
            ATTR_USER_NAME,
            group=GROUP_RC,
            default_value=None,
            readonly=True,
            display=False,
        )
        self.create_attribute(
            ATTR_PASSWORD,
            group=GROUP_RC,
            default_value=None,
            readonly=True,
            display=False,
        )
        self.create_attribute(
            ATTR_RC_MACRO_DISCUSSION, group=GROUP_RC, default_value=None, readonly=True
        )
        self.create_attribute(
            ATTR_HOST,
            group=GROUP_RC,
            default_value="https://cfel-rocketchat.desy.de",
            readonly=True,
        )
        self.create_attribute(
            ATTR_LOOPS_RUNNING, default_value=False, default_type=bool, readonly=True
        )
        self.create_attribute(ATTR_PROCESS_ID, default_value=os.getpid(), readonly=True)
        self.create_attribute(
            ATTR_CPU_USAGE,
            default_value=0,
            default_type=float,
            readonly=True,
            unit="%",
            decimals=4,
        )
        self.create_attribute(
            ATTR_MEMORY_USAGE,
            default_value=0,
            default_type=float,
            readonly=True,
            decimals=3,
            unit="MB",
            factor=10e-7,
        )
        self.create_attribute(
            ATTR_AVAILABLE_MEMORY,
            default_value=0,
            default_type=float,
            readonly=True,
            unit="MB",
            factor=10e-7,
            decimals=3,
        )
        self.set_raw_value(ATTR_AVAILABLE_MEMORY, psutil.virtual_memory().total)

    def _config_attributes_log(self):
        """
        Config attribute log.
        Device and Attribute must be defined in Session config under logged_attributes key.
        """
        if self.config:
            logger = self.get_device("AttributeLogger")
            for device_id, attribute in self.config.get(
                "logged_attributes", {}
            ).items():
                if isinstance(attribute, list):
                    for sub_attribute in attribute:
                        assert isinstance(logger, AttributeLogger)
                        logger.add_logged_attribute(device_id, sub_attribute)
                else:
                    assert isinstance(logger, AttributeLogger)
                    logger.add_logged_attribute(device_id, attribute)
            logging_interval = self.config.get("logging_interval", False)
            if logging_interval:
                logger.set_value(ATTR_INTERVAL, logging_interval)
            logger.set_attribute((ATTR_LOGGING, VALUE), True)

    def handle_configuration(self):
        """
        Handle Session configuration.
        We want to set Loggers and initiate cache directory here.
        """
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()

            if self.get_value(ATTR_ALLOW_CONSOLE_LOG):
                self.set_console_log_directory("Console")
            if self.get_value(ATTR_ALLOW_GUI_LOG):
                self.set_gui_log_directory("Gui")
            if self.get_value(ATTR_ALLOW_DEVICE_LOG):
                self.set_device_log_directory("Device")
            if self.get_value(ATTR_ALLOW_ATTRIBUTE_LOG):
                self.set_attribute_log_directory("Attributes")
                self._config_attributes_log()
            if self.get_value(ATTR_ALLOW_MACRO_LOG):
                self.set_macro_log_directory("Macro")
            if self.get_value(ATTR_CACHE_DIRECTORY) is None:
                self.set_value(ATTR_CACHE_DIRECTORY, ".cache")

            # Start polling of CPU and RAM usage
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(f"Device configuration took {time.time() - start_at} sec.")

        _finish_configuration()

    def add_measurement_group(self, measurement_group: MeasurementGroup):
        """
        Add new measurement group into the session.
        Measurement group is used to provide group of Device Attributes, which are interesting.
        This groups then can be used for logging or any other case.
        :param MeasurementGroup measurement_group: new measurement group
        :return:
        """
        assert isinstance(measurement_group, MeasurementGroup)
        self.measurement_groups[measurement_group.group_id] = measurement_group

    # pylint: disable=unused-argument
    def set_session(self, session: "DeviceSession") -> None:
        """
        Even Session has to have a reference to Session.
        In this case it reference to itself.
        :param DeviceSession session: Session object
        :return:
        """
        self.devices[self.device_id] = self
        self.session = self
        self.set_value(ATTR_DEVICES_COUNT, len(self.devices))

    def set_config(self, config_file: str):
        """
        Set config file for a Session.
        It's used for the case of config update and write.
        It set session_config variable as a tuple (config_file, config dict).
        Use set_config in the script that start the session (server.py, client.py).
        :param str config_file: Path to the configuration file
        """
        with self.config_lock:
            # pylint: disable=unspecified-encoding
            with open(config_file, "r") as configFile:
                self.session_config = (
                    config_file,
                    yaml.load(configFile, Loader=YamlConfigSerializer),
                )

    def update_config(
        self, device_id: str, attribute: Union[List, Tuple], value: object
    ):
        """
        When Device Attribute has save_change set to True, we want to update config and save it.
        This can be done only when session_config was set.
        :param str device_id: Device ID
        :param Attribute attribute: Attribute to update
        :param object value: Value of the Attribute
        """
        with self.config_lock:
            if self.session_config is None:
                return
            else:
                config_file, parsed_config = self.session_config
                device_config = parsed_config["devices"][device_id]
                if "config" not in device_config.conf:
                    device_config.conf["config"] = {
                        "attributes": {tuple(attribute): value}
                    }
                elif "attributes" not in device_config.conf["config"]:
                    device_config.conf["config"]["attributes"] = {
                        tuple(attribute): value
                    }
                else:
                    device_config.conf["config"]["attributes"][tuple(attribute)] = value
                # pylint: disable=unspecified-encoding
                with open(config_file, "w") as fp:
                    yaml.dump(parsed_config, fp, Dumper=YamlDumper)

    def command(self, command, callback=None, with_token=False, returning=True) -> int:
        """
        Session is Device, we have to disable command option.
        :param str command: Request to Device
        :param Callable callback: callback after Device replies to Command
        :param bool with_token: token
        :param returning: bool
        :return int: token
        """
        raise DeviceError("Device does not accept any commands.")

    def get_device(self, device_id: str, search_master: bool = True):
        """
        Get Device object from the session.
        If master_server key is defined in config, Session will try to obtain Device from it.
        If Device is not found, raise DeviceUnknownError exception.
        :param str device_id: Device ID
        :param bool search_master: Ask master server for Device
        :raises DeviceUnknownError
        :return Device: Device object
        """
        try:
            return self.devices[device_id]
        except KeyError:
            if search_master:
                master_server, master_port = (
                    self.config.get("master_server", False),
                    self.config.get("master_port", False),
                )
                if master_server and master_port:
                    return DeviceClient(master_server, master_port, device_id)
            raise DeviceUnknownError(f"Device {device_id} was not found.")

    def add_device(self, device: Device):
        """
        Add Device into the Session
        Raise DeviceError when Device is already registered.
        :param Device device: Device object
        :raises DeviceError
        """
        assert isinstance(device, Device)

        if device.device_id not in self.devices:
            self.devices[device.device_id] = device
            device.session = self
            self.set_value(ATTR_DEVICES_COUNT, len(self.devices))
        else:
            raise DeviceError(
                f"Device {device.device_id} is already registered within session."
            )

    def remove_device(self, device: Device):
        """
        Remove Device from the Session.
        If Device does not exists, not raise any Exception, this is on purpose :).
        :param Device device: Device object
        """
        assert isinstance(device, Device)
        if device.device_id in self.devices:
            device.session = None
            del self.devices[device.device_id]
            self.set_value(ATTR_DEVICES_COUNT, len(self.devices))

    def get_devices(
        self, class_filter=None, attribute_filter=None, method_filter=None
    ) -> OrderedDict[str, Device]:
        """
        Filter devices in current session.

        :param list class_filter: list of classes
        :param attribute_filter: attribute
        :param method_filter: method name
        :return list: list of filtered devices
        """
        filtered_devices = self.devices.copy()
        # Filter over device class
        if class_filter is not None:
            for device_id, device in filtered_devices.copy().items():
                if class_filter not in device.qualified_name:
                    del filtered_devices[device_id]
        # Filter over device attributes
        if attribute_filter is not None:
            for device_id, device in filtered_devices.copy().items():
                if device.get_attribute(attribute_filter) is None:
                    del filtered_devices[device_id]
        # Filter over device methods
        if method_filter is not None:
            for device_id, device in filtered_devices.copy().items():
                filtered_out_methods = filter(
                    lambda val: val[0] == method_filter, device.exposed_methods
                )
                try:
                    next(filtered_out_methods)
                except StopIteration:
                    del filtered_devices[device_id]

        return filtered_devices

    def change_device_id(self, device_id, new_device_id):
        """
        Change ID of Device in Session
        :param str device_id: old Device ID
        :param str new_device_id: new Device ID
        """
        device = self.get_device(device_id)
        device.device_id = new_device_id
        self.devices[new_device_id] = self.devices.pop(device_id)

    def set_log_directory(self, value: str):
        """
        Set main log Directory.
        Try to create it, if it does not exist.
        :param str value: Directory path
        """
        self.logger.info(f"Setting log directory to: {value}")
        if not os.path.exists(value):
            self.logger.info(f"Directory {value} does not exists, trying to create it.")
            os.makedirs(value)

    def set_resource_directory(self, value: str):
        """
        Set main resource Directory.
        Try to create it, if it does not exist.
        :param str value: Resource path
        """
        self.logger.info(f"Setting resource directory to: {value}")
        if not os.path.exists(value):
            self.logger.info(f"Directory {value} does not exists, trying to create it.")
            os.makedirs(value)

    def set_cache_directory(self, value: str):
        """
        Set main cache Directory.
        Try to create it, if it does not exist.
        :param str value: Cache path
        """
        self.logger.info("Setting cache directory to: {}".format(value))
        if not os.path.exists(value):
            self.logger.info(f"Directory {value} does not exists, trying to create it.")
            os.makedirs(value)

    def set_console_log_directory(self, value: str):
        """
        Set console log Directory.
        Try to create it, if it does not exist.
        :param str value: Console log Directory path
        :raises DeviceError
        """
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(
                "Cannot set console log directory. Attribute 'Log directory' is not set."
            )
        log_directory = os.path.join(log_directory, value)
        self.logger.info(f"Setting console log directory to: {log_directory}")
        if not os.path.exists(log_directory):
            self.logger.info(
                f"Directory {log_directory} does not exists, trying to create it."
            )
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger("Console")
        set_rotating_file_handler(
            logger, os.path.join(log_directory, self.LOG_CONSOLE_OUTPUT_FILE)
        )
        set_sys_exception_handler(logger)
        logger.setLevel(self.log_level)
        logger.addHandler(handler)

    def set_gui_log_directory(self, value: str):
        """
        Set gui log Directory.
        Try to create it, if it does not exist.
        :param str value: Gui log Directory path
        :raises DeviceError
        """
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(
                "Cannot set gui log directory. Attribute 'Log directory' is not set."
            )
        log_directory = os.path.join(log_directory, value)

        self.logger.info(f"Setting gui log directory to: {log_directory}")
        if not os.path.exists(log_directory):
            self.logger.info(
                f"Directory {log_directory} does not exists, trying to create it."
            )
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger("Gui")
        set_rotating_file_handler(
            logger, os.path.join(log_directory, self.LOG_GUI_OUTPUT_FILE)
        )
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    def set_attribute_log_directory(self, value: str):
        """
        Set Attribute log Directory.
        Try to create it, if it does not exist.
        :param str value: Attribute log Directory path
        """
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(
                "Cannot set Attribute log directory. Attribute 'Log directory' is not set."
            )
        log_directory = os.path.join(log_directory, value)

        self.logger.info("Setting device log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(
                "Directory {} does not exists, trying to create it.".format(
                    log_directory
                )
            )
            os.makedirs(log_directory)

        logger = AttributeLogger(
            os.path.join(log_directory, self.LOG_ATTRIBUTES_OUTPUT_FILE),
            device_id="AttributeLogger",
        )
        self.attributes.update({"Attribute logger": logger.attributes})

    def set_device_log_directory(self, value):
        """
        Set device log Directory.
        Try to create it, if it does not exist.
        :param str value: Device log Directory path
        :raises DeviceError
        """
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(
                "Cannot set device log directory. Attribute 'Log directory' is not set."
            )
        log_directory = os.path.join(log_directory, value)

        self.logger.info("Setting device log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(
                "Directory {} does not exists, trying to create it.".format(
                    log_directory
                )
            )
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger("Device")
        set_rotating_file_handler(
            logger, os.path.join(log_directory, self.LOG_DEVICE_OUTPUT_FILE)
        )
        logger.setLevel(self.log_level)
        logger.addHandler(handler)

    def set_macro_log_directory(self, value):
        """
        Set Macro log Directory.
        Try to create it, if it does not exist.
        :param str value: Macro log Directory path
        :raises DeviceError
        """
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(
                "Cannot set macro log directory. Attribute 'Log directory' is not set."
            )
        log_directory = os.path.join(log_directory, value)

        self.logger.info(f"Setting gui log directory to: {log_directory}")
        if not os.path.exists(log_directory):
            self.logger.info(
                f"Directory {log_directory} does not exists, trying to create it."
            )
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger("Macro")
        set_rotating_file_handler(
            logger, os.path.join(log_directory, self.LOG_MACRO_OUTPUT_FILE)
        )
        if self.get_value([GROUP_RC, ATTR_ALLOW_RC]):
            discussion_name = self.get_value([GROUP_RC, ATTR_RC_MACRO_DISCUSSION])
            if discussion_name is None:
                discussion_name = "{}-{}".format(self.device_id.lower(), "macro-log")
                self.set_value([GROUP_RC, ATTR_RC_MACRO_DISCUSSION], discussion_name)
            user_credentials, auth_token = self._get_rc_auth()
            set_rocket_chat_handler(
                logger,
                level=logging.MACRO_STATE,
                user_credentials=user_credentials,
                auth_token=auth_token,
                main_room=self.device_id.lower(),
                discussion=discussion_name,
            )
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    def _get_rc_auth(self):
        """
        Helper method to get RocketChat credentials.
        :return tuple: ((RocketChat user, RocketChat password), RocketChat auth token)
        """
        rc_user = self.get_value([GROUP_RC, ATTR_USER_NAME])
        rc_pass = self.get_value([GROUP_RC, ATTR_PASSWORD])
        rc_auth_token = self.get_value([GROUP_RC, ATTR_AUTH_TOKEN])
        return (rc_user, rc_pass), rc_auth_token

    def start_control_loops(self):
        if not self.get_value(ATTR_LOOPS_RUNNING):
            self.start_asyncore_loop()
            self.start_device_connection_poller()
            self.start_device_poller_loop()
            self.start_port_read_loop()
            self.set_value(ATTR_LOOPS_RUNNING, True)

    def stop_control_loops(self):
        if self.get_value(ATTR_LOOPS_RUNNING):
            self.stop_asyncore_loop()
            self.stop_port_read_loop()
            self.stop_device_poller_loop()
            self.stop_device_connection_poller()
            self.set_value(ATTR_LOOPS_RUNNING, False)

        self.logger.info("Closing all control loops")
        # Here we try to join all control loops giving them 1 second timeout
        control_asyncore_loop.join(1)
        control_device_poller.join(1)
        control_port_read_loop.join(1)
        control_device_connection_poller.join(1)

        self.logger.info("Control loops closed")

    def stop(self):
        """Stop and disconnect all devices in Session."""
        self.stop_polling()
        if self.get_value(ATTR_ALLOW_ATTRIBUTE_LOG):
            logger = self.get_device("AttributeLogger", search_master=False)
            if logger is not None:
                assert isinstance(logger, AttributeLogger)
                logger.stop()

        if self.server is not None:
            self.server.stop()

        if self.publisher is not None:
            self.publisher.stop()

        self.session = None
        del self.devices[self.device_id]

        for device in list(self.devices.values()):
            device.disconnect()

        if self.get_value(ATTR_LOOPS_RUNNING):
            self.stop_control_loops()
            self.set_value(ATTR_LOOPS_RUNNING, False)

    @staticmethod
    def start_asyncore_loop():
        control_asyncore_loop.start()

    @staticmethod
    def stop_asyncore_loop():
        control_asyncore_loop.stop()

    @staticmethod
    def start_device_poller_loop():
        control_device_poller.start()

    @staticmethod
    def stop_device_poller_loop():
        control_device_poller.stop()

    @staticmethod
    def start_port_read_loop():
        control_port_read_loop.start()

    @staticmethod
    def stop_port_read_loop():
        control_port_read_loop.stop()

    @staticmethod
    def start_device_connection_poller():
        control_device_connection_poller.start()

    @staticmethod
    def stop_device_connection_poller():
        control_device_connection_poller.stop()

    def set_high_process_priority(self):
        """
        Set high process affinity.

        In Windows use win32process.REALTIME_PRIORITY_CLASS, for Linux change nice to -20.
        """
        if psutil.WINDOWS:
            try:
                import win32api
                import win32process
                import win32con
            except ImportError:
                self.logger.error("Package pypiwin32 is not installed")
            pid = win32api.GetCurrentProcessId()
            handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
            win32process.SetPriorityClass(handle, win32process.REALTIME_PRIORITY_CLASS)
            # win32process.SetThreadPriority(win32api.GetCurrentThread(), win32process.THREAD_PRIORITY_TIME_CRITICAL)
        else:
            try:
                process = psutil.Process(os.getpid())
                process.nice(-20)
            except psutil.AccessDenied:
                self.logger.error("Could not raise process priority")

    @expose_method()
    def clear_cache_folder(self):
        """Clear all files from current '.cache' folder."""
        folder = self.get_value(ATTR_CACHE_DIRECTORY)
        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                self.logger.error(f"Failed to delete {file_path}: {e}")

    def measure_cpu_ram(self):
        self.set_raw_value(ATTR_CPU_USAGE, self.os_process.cpu_percent())
        self.set_raw_value(ATTR_MEMORY_USAGE, self.os_process.memory_info().rss)

    def start_polling(self):
        self.polling_timer.start()

    def stop_polling(self):
        self.polling_timer.stop()


def server_process(conf: str, stop_event: EventType) -> Process:
    """
    Here comes a short description of server_process.

    For the reason of having multiple Processes handling multiple devices, we can start
    each server in it's own Process. This function will start Server in new Process
    and return that Process. If stop event is set, then stop session.

    :param conf: path to the config file
    :param stop_event: event
    :return: Process
    """

    def _process(conf, stop_event):
        # I have no idea why this is reimported, but maybe it makes sense?
        # pylint: disable=redefined-outer-name,reimported
        import kamzik3

        # pylint: disable=redefined-outer-name
        import yaml
        from time import sleep

        # pylint: disable=unspecified-encoding
        with open(conf, "r") as configFile:
            yaml.load(configFile, Loader=yaml.Loader)

        # Start control loop for devices specified in configuration file
        kamzik3.session.set_config(conf)
        kamzik3.session.start_control_loops()
        while not stop_event.is_set():
            try:
                sleep(1)
            except KeyboardInterrupt:
                break
        kamzik3.session.stop()

    return Process(target=_process, args=(conf, stop_event))



import random
import time
from threading import Thread
from time import sleep

import numpy
import pandas

from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.devices.dummy.deviceDummy import DeviceDummy
from kamzik3.snippets.snippetsDecorators import expose_method


class DummyInficonQcm(DeviceDummy):
    def __init__(self, device_id=None, config=None):
        self._materials = pandas.read_csv(
            "./resources/inficon_qcm_materials.txt", sep="\t", names=["Formula", "Name"]
        )
        self.materials = (
            self._materials["Formula"] + " - " + self._materials["Name"]
        ).to_list()
        DeviceDummy.__init__(self, device_id, config)

    def _init_attributes(self):
        DeviceSocket._init_attributes(self)
        self.create_attribute(ATTR_PRODUCT_TYPE, readonly=True)
        self.create_attribute(ATTR_SOFTWARE_VERSION, readonly=True)
        self.create_attribute(ATTR_FIRMWARE, readonly=True)
        self.create_attribute(ATTR_SENSORS_COUNT, default_value=8, readonly=True)
        self.create_attribute(
            ATTR_ACTIVE_PROCESS, default_value=1, default_type=int, readonly=True
        )

        for i in range(1, self.get_value(ATTR_SENSORS_COUNT) + 1):
            self.create_attribute(
                f"{ATTR_MATERIAL}_{i}",
                default_value=None,
                default_type=self.materials,
                group=GROUP_MATERIALS,
            )
            self.create_attribute(
                f"{ATTR_OUTPUT}_{i}", default_value=None, group=GROUP_OUTPUTS
            )
            self.create_attribute(
                f"{ATTR_PROCESS}_{i}", default_value=None, group=GROUP_PROCESSES
            )
            self.create_attribute(
                f"{ATTR_USER_MESSAGE}_{i}",
                default_value=None,
                group=GROUP_USER_MESSAGES,
            )

    @expose_method()
    def init_measurement(self):
        active_channel = self.get_value(ATTR_ACTIVE_PROCESS) - 1
        self.notify((active_channel, ATTR_LAYER_THICKNESS), 0)
        self.notify((active_channel, ATTR_SENSOR_THICKNESS), 0)

    @expose_method()
    def stop_measurement(self):
        pass

    def set_active_process(self, process):
        self.set_value(ATTR_ACTIVE_PROCESS, process)

    def notify_channels(self, attribute, value):
        for channel in range(self.get_value(ATTR_SENSORS_COUNT)):
            self.notify((channel, attribute), value)


class DummyInficonQcmSensorChannel(DeviceChannel):
    sensor_state = [
        "Good Crystal and active",
        "Failed Crystal and active",
        "Invalid Measurement on Crystal and active",
        "Good Crystal and inactive",
        "Failed Crystal and inactive",
        "Invalid Measurement on Crystal and inactive",
    ]
    sensor_type = ["Single", "XtalTwo", "XtalSix", "Xtal12", "Generic"]

    def __init__(self, device, channel, device_id=None, config=None):
        super().__init__(device, channel, device_id, config)
        self.handle_configuration()

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            Thread(target=self._sensor_simulation).start()
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        self.connected = True
        _finish_configuration()

    def _init_attributes(self):
        DeviceChannel._init_attributes(self)
        self.create_attribute(ATTR_STATE, readonly=True)
        self.add_attribute(
            ATTR_MATERIAL,
            self.device.get_attribute(
                [GROUP_MATERIALS, f"{ATTR_MATERIAL}_{self.channel + 1}"]
            ),
        )
        self.create_attribute(ATTR_SENSOR_TYPE, default_type=self.sensor_type)
        self.create_attribute(ATTR_AUTO_Z, default_type=bool)
        self.create_attribute(
            ATTR_Z_RATIO, default_value=0, default_type=float, readonly=True, decimals=9
        )
        self.create_attribute(
            ATTR_FUNDAMENTAL_FREQUENCY,
            unit="Hz",
            default_value=0,
            default_type=numpy.double,
            readonly=True,
            decimals=9,
            factor=0.000873114913702011,
        )
        self.create_attribute(
            ATTR_AHARMONIC_FREQUENCY,
            unit="Hz",
            default_value=0,
            default_type=numpy.double,
            readonly=True,
            decimals=9,
            factor=0.000873114913702011,
        )
        self.create_attribute(
            ATTR_SENSOR_THICKNESS,
            unit="nm",
            default_value=0,
            default_type=float,
            readonly=True,
            decimals=9,
            factor=0.1,
        )
        self.create_attribute(
            ATTR_CRYSTAL_LIFE,
            unit="%",
            default_value=100,
            default_type=int,
            readonly=True,
        )
        self.create_attribute(
            ATTR_RAW_RATE,
            unit="nm/s",
            default_value=0,
            default_type=float,
            readonly=True,
            decimals=9,
            factor=0.1,
        )
        self.create_attribute(
            ATTR_SHUTTER_STROKES,
            default_value=0,
            default_type=int,
            readonly=True,
            save_change=True,
            min_value=0,
        )
        self.create_attribute(
            ATTR_SHUTTER_OPEN, default_value=False, default_type=bool, readonly=True
        )
        self.create_attribute(
            ATTR_LAYER_THICKNESS,
            unit="nm",
            default_value=0,
            default_type=float,
            readonly=True,
            decimals=9,
            factor=0.1,
        )

    def _sensor_simulation(self):
        while self.configured:
            factor = 1000 if self.get_value(ATTR_SHUTTER_OPEN) else 1
            lt = self.get_value(ATTR_LAYER_THICKNESS)
            st = self.get_value(ATTR_SENSOR_THICKNESS)
            self.set_value(
                ATTR_AHARMONIC_FREQUENCY, random.randint(10000, 100000) * factor
            )
            self.set_value(
                ATTR_FUNDAMENTAL_FREQUENCY, random.randint(10000, 100000) * factor
            )
            raw_rate = (random.randint(1, 1000) * 1e-9) * factor
            self.set_value(ATTR_RAW_RATE, raw_rate)
            if self.device.get_value(ATTR_ACTIVE_PROCESS) == self.channel + 1:
                self.set_value(ATTR_LAYER_THICKNESS, lt + raw_rate)
            self.set_value(ATTR_SENSOR_THICKNESS, st + raw_rate)
            self.set_value(ATTR_CRYSTAL_LIFE, 100 - (lt + raw_rate))
            sleep(0.1)

    def subject_update(self, key, value, subject):
        DeviceChannel.subject_update(self, key, value, subject)

        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]

            self.set_attribute(
                (ATTR_LATENCY, VALUE), self.device.get_value(ATTR_LATENCY)
            )
            self.set_value(attribute, value)

    @expose_method()
    def open_shutter(self):
        if not self.get_value(ATTR_SHUTTER_OPEN):
            self.set_value(
                ATTR_SHUTTER_STROKES, self.get_value(ATTR_SHUTTER_STROKES) + 1
            )
            self.set_value(ATTR_SHUTTER_OPEN, True)

    @expose_method()
    def close_shutter(self):
        if self.get_value(ATTR_SHUTTER_OPEN):
            self.set_value(
                ATTR_SHUTTER_STROKES, self.get_value(ATTR_SHUTTER_STROKES) + 1
            )
            self.set_value(ATTR_SHUTTER_OPEN, False)

    @expose_method()
    def measure_layer_thickness(self):
        self.device.set_active_process(self.channel + 1)

    @expose_method()
    def clear_shutter_strokes(self):
        self.set_value(ATTR_SHUTTER_STROKES, 0)



import logging
import time
from collections import deque
from copy import copy
from math import inf
from threading import Thread
from typing import Any, Callable, Tuple, List, Optional, Union, TYPE_CHECKING

import numpy as np
import oyaml as yaml
from pint import UndefinedUnitError, DimensionalityError
import serial
import warnings

import kamzik3
from kamzik3 import WriteException, CommandFormatException, DeviceError
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute, SetFunction
from kamzik3.devices.subject import Subject
from kamzik3.snippets.snippetDataAccess import get_from_dict, fullname, is_equal
from kamzik3.snippets.snippetsControlLoops import (
    control_device_poller,
    control_device_connection_poller,
)
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsGenerators import device_id_generator, token_generator
from kamzik3.snippets.snippetsUnits import device_units
from kamzik3.snippets.snippetsYaml import YamlSerializable

if TYPE_CHECKING:
    from kamzik3.devices.deviceSession import DeviceSession


class Device(Subject, YamlSerializable):
    """
    Device is a main parent class for any Device within Kamzik3.
    It provides methods to work with Attributes, Connection and Commands handling.
    If You inherit from Device, You must reimplement handle_configuration method.

    Connection is done in sequence:
        - connect()
        - handle_connect_event()
            - handle_connect()
            - handle_configuration_event()
                - handle_configuration()

    Connection error handling:
        - if connect() => error state
           - disconnect() if critical error
           - close(), reconnect() if not critical error
        - if connect() => connection timeout
            - handle_connection_error(), close()
        - if Device.connected and response timeout()
            - handle_response_error(), close_connection()

    Disconnection is done in sequence:
        - disconnect()
            - close() if not Device.connected
            - close_connection() if connected
                - close()

    General config setup in Yaml example:

    # List of devices
    devices:
        # Device_id_within_yml: Kamzik3 absolute path to Device class
        MyMotorX: !Device:kamzik3.devices.dummy.dummyMotor.DummyMotor
            device_id: MyMotorX
            Motor1_axis_0: &x !Device:kamzik3.devices.dummy.dummyMotor.DummyMotorChannel
            channel: 0
            config:
                attributes:
                    !!python/tuple [Velocity, Value]: 1mm/s
                    !!python/tuple [Position, Tolerance]: [100nm, 100nm]
        # Device 2 definition
        # Device 3 definition
        # ...
    """

    id_generator = device_id_generator()  # Device id generator
    connection_timeout = 4000  # in ms
    response_timeout = 4000  # in ms
    session: Optional["DeviceSession"] = None  # Holds current Session object
    push_commands_max = inf  # Max commands, that Device can handle at once
    push_buffer_size = 2**16  # Max number of characters per request
    command_encoding: Optional[str] = "utf8"  # Default command encoding

    def __init__(self, device_id=None, config=None):
        self.connecting_time = 0  # Counting time of connection
        self.request_timestamp = 0  # Timestamp of last request
        self.response_timestamp = 0  # Timestamp of last Device response to the request
        # Connection flags
        self.connected = False
        self.connecting = False
        self.closing = False
        self.closed = False
        # Connection error flags
        self.connection_error = False
        self.response_error = False
        self.latency_buffer = deque(
            maxlen=20
        )  # response - request time, calculate avg response latency
        self.init_time = time.time()  # Initiation time
        self.commands_buffer = deque()  # Holds requested commands
        Subject.__init__(self)

        self.device_id = device_id
        if device_id is None:
            # If no device_id was not specify, generate one
            self.device_id = next(Device.id_generator)
        self.config = config
        if self.config is None:
            # Config must be dictionary
            self.config = {}
        self.token_generator = token_generator()
        self.device_poller = control_device_poller  # Attach poller loop
        self.device_connection_poller = (
            control_device_connection_poller  # Attach connection loop
        )
        self.qualified_name = fullname(self)  # Get absolute path to the Device class
        if not hasattr(self, "logger"):
            # Attach Device log, if it does not exists
            self.logger = logging.getLogger("Device.{}".format(self.device_id))
        if "push_commands_max" in self.config:
            # Check and set push_commands_max from Config file
            self.push_commands_max = self.config["push_commands_max"]
        if "push_buffer_size" in self.config:
            # Check and set push_buffer_size from Config file
            self.push_buffer_size = self.config.get("push_buffer_size")
        # Filter for Attributes and Methods, which can be used within Macro
        # By Default use all attributes and methods
        self.macro_steps = {
            MACRO_SET_ATTRIBUTE_STEP: ["*"],
            MACRO_EXECUTE_METHOD_STEP: ["*"],
        }
        self.exposed_methods = []  # List of exposed methods
        self._expose_methods_to_clients()
        if not hasattr(self, "attributes"):
            # Check if attributes are already defined
            self.attributes = {}
            self.attributes_sharing_map = {}
            self.attribute_attach_map = {}
            self._init_attributes()

        if kamzik3.session is not None:
            # Check if device id is already registered in session
            # and remove it if it's not our current Device
            if (
                self.device_id in kamzik3.session.devices
                and self != kamzik3.session.devices[self.device_id]
            ):
                kamzik3.session.devices[self.device_id].close()
                kamzik3.session.devices[self.device_id] = None
                del kamzik3.session.devices[self.device_id]

            if self.device_id not in kamzik3.session.devices:
                # Check if device is not registered in session and set current session
                self.set_session(kamzik3.session)

    def __getitem__(self, key):
        """
        To get Device's Attribute we can use directly dict key notation.
        For examples self[(POSITION, VALUE)] gets POSITION Value.

        :param key: Get Attribute from Device.attributes dict
        :return: Attribute or Value
        """
        return self.attributes[key]

    def __setitem__(self, key, value):
        """
        To set Device's Attribute we can use directly dict key notation.
        For examples self[(POSITION, VALUE)] = 50nm

        Be careful not to use self[POSITION] = 50m, this would set Attribute to just
        50nm.

        :param key: Set Attribute Value
        :param value: mixed, based Attribute's TYPE
        """
        self.attributes[key] = value

    def _init_attributes(self):
        """
        Initiate Device's attributes.

        To call this method is mandatory.
        Overload this method for any other derived Device.
        Attributes;
            ATTR_ID: Unique Device ID
            ATTR_STATUS: Current Device status can be user defined, but it's overwritten
            by connection methods
            ATTR_DESCRIPTION: Attribute's description, it's displayed in GUI on
            MouseHover
            ATTR_ENABLED: Meta flag for GUI to Enable / Disable edit of Value
            ATTR_LATENCY: Average Device response time to request
            ATTR_BUFFERED_COMMANDS: Number of buffered commands. Higher number can
            indicates ERROR.
            ATTR_HANGING_COMMANDS: Number of commands sent to Device and waiting
            response. Higher number can indicates ERROR.
            ATTR_LAST_ERROR: Last occurred error
        """

        def set_status(key, value):
            """
            When status has changed, notify all Observers
            :param key: str
            :param value: str
            """
            if key == VALUE:
                self.notify(ATTR_STATUS, value)

        self.create_attribute(
            ATTR_ID,
            default_value=self.device_id,
            readonly=True,
            description="Unique device ID",
        )
        self.create_attribute(
            ATTR_STATUS,
            default_value=STATUS_DISCONNECTED,
            readonly=True,
            description="Current device status",
            min_broadcast_timeout=0,
        ).attach_callback(set_status)
        self.create_attribute(
            ATTR_DESCRIPTION, readonly=True, description="Device description"
        )
        self.create_attribute(
            ATTR_ENABLED,
            default_value=True,
            default_type=bool,
            description="Allow any external changes to device",
        )
        self.create_attribute(
            ATTR_LATENCY,
            default_value=0,
            default_type=np.uint16,
            description="Latency between sending command and receiving an answer",
            min_value=0,
            max_value=9999,
            unit="ms",
            readonly=True,
        )
        self.create_attribute(
            ATTR_BUFFERED_COMMANDS,
            default_value=0,
            default_type=np.uint32,
            description="Amount of commands waiting to be executed",
            min_value=0,
            readonly=True,
        )
        self.create_attribute(
            ATTR_HANGING_COMMANDS,
            default_value=0,
            default_type=np.uint32,
            description="Amount of commands waiting to be answered from device",
            min_value=0,
            readonly=True,
        )
        self.create_attribute(
            ATTR_LAST_ERROR,
            readonly=True,
            description="Last device error exception",
            min_broadcast_timeout=0,
        )

    def create_attribute(self, name, group=None, **kwargs) -> Attribute:
        """
        Use this function to create new attribute.
        !!! It's mandatory to use ONLY this method to create new Device's Attributes !!!
        :param str name: Name of the Attributes
        :param str group: Group of the Attribute
        :param mixed kwargs: Defined key:value pairs of the Attribute
        :return: Attribute
        """
        attribute_path = name
        if group is not None:
            attribute_path = "{}.{}".format(group, attribute_path)
        attribute_id = "{}.{}.{}".format(
            self.device_id, TOKEN_ATTRIBUTE, attribute_path
        )
        attribute = Attribute(attribute_id, **kwargs)

        if group is None:
            self.attributes[name] = attribute
        else:
            if group not in self.attributes:
                self.attributes[group] = {}
            self.attributes[group][name] = attribute
        if kamzik3.session is not None and kamzik3.session.publisher is not None:
            header = "{}.{}".format(attribute.attribute_id, TOKEN_ATTRIBUTE_REPLACE)
            kamzik3.session.publisher.push_message(header, attribute)
        return attribute

    def add_attribute(
        self, name: str, attribute: Attribute, group: Optional[str] = None
    ) -> Attribute:
        """
        Use this function to add existing attribute to Device.

        This is useful if Attribute from other Device has to be visible in this Device
        as well.

        Use only create_attribute when you need to add new Attribute to this Device.
        Set group parameter if You want attribute be in specific group.

        :param name: str
        :param attribute: Attribute
        :param group: str
        :return: Attribute
        """
        attribute_list = [name]
        if group is None:
            self.attributes[name] = attribute
        else:
            attribute_list.insert(0, group)
            if group not in self.attributes:
                self.attributes[group] = {}
            self.attributes[group][name] = attribute

        attribute_exists = attribute.attribute_id in self.attributes_sharing_map
        if attribute_exists:
            self.attributes_sharing_map[attribute.attribute_id].append(attribute_list)
        else:
            self.attributes_sharing_map[attribute.attribute_id] = [attribute_list]
        if (
            attribute_exists
            and kamzik3.session is not None
            and kamzik3.session.publisher is not None
        ):
            header = "{}.{}".format(attribute.attribute_id, TOKEN_ATTRIBUTE_REPLACE)
            kamzik3.session.publisher.push_message(header, attribute)
        if attribute.attribute_id is None:
            attribute_path = name
            if group is not None:
                attribute_path = "{}.{}".format(group, attribute_path)
            attribute.attribute_id = "{}.{}.{}".format(
                self.device_id, TOKEN_ATTRIBUTE, attribute_path
            )
        return attribute

    def delete_attribute(self, name: str, group: Optional[str] = None) -> None:
        """
        Delete specific attribute.

        :param name: Name of the Attribute
        :param group: Optional group of the Attribute
        """
        attribute_path = name
        if group is not None:
            attribute_path = "{}.{}".format(group, attribute_path)
        attribute_id = "{}.{}.{}".format(
            self.device_id, TOKEN_ATTRIBUTE, attribute_path
        )
        if group is not None:
            del self.attributes[group][name]
        else:
            del self.attributes[name]
        if kamzik3.session is not None and kamzik3.session.publisher is not None:
            header = "{}.{}".format(attribute_id, TOKEN_ATTRIBUTE_DELETE)
            kamzik3.session.publisher.push_message(header, attribute_path)

    def delete_attribute_group(self, group: str) -> None:
        """
        Delete all attributes including group.

        :param str group: Group of the Attribute
        """
        attribute_path = group
        attribute_id = "{}.{}.{}".format(
            self.device_id, TOKEN_ATTRIBUTE, attribute_path
        )
        del self.attributes[group]
        if kamzik3.session is not None and kamzik3.session.publisher is not None:
            header = "{}.{}".format(attribute_id, TOKEN_ATTRIBUTE_GROUP_DELETE)
            kamzik3.session.publisher.push_message(header, attribute_path)

    def share_group(
        self, source_device, source_group, target_group, attribute_name_mask=None
    ):
        """
        Share all attributes in group.

        Filter by attribute_name_mask from source_device.

        :param Device source_device: Source Device object
        :param str source_group: Name of the group to share
        :param str target_group: Target group on this Device
        :param dict attribute_name_mask: Use this, if You need to rename Attribute
        """
        if attribute_name_mask is None:
            attribute_name_mask = {}
        if source_group is not None:
            from_group = source_device.attributes[source_group]
        else:
            from_group = source_device.attributes

        for attribute_name, attribute in from_group.items():
            self.add_attribute(
                attribute_name_mask.get(attribute_name, attribute_name),
                attribute,
                target_group,
            )

    def share_exposed_method(
        self,
        source_device,
        source_method,
        shared_method_name=None,
        shared_method_attributes=None,
    ):
        """
        Share exposed method with source_device.

        This is handy method when you share attributes of other device and you want
        to also expose it's methods. Each new method will go under new name
        source_device.id_source_method.name. If You define shared_method_attributes
        they will overwrite default ones.

        :param Device source_device: Source Device object
        :param source_method: Name of source exposed method
        :param str shared_method_name: optional new name of the method
        :param shared_method_attributes: optional Dictionary of method attributes
        """
        for method_name, method_parameters in source_device.exposed_methods:
            if method_name == source_method:
                if shared_method_name is None:
                    shared_method_name = "{}_{}".format(
                        source_device.device_id, method_name
                    )
                if shared_method_attributes is None:
                    shared_method_attributes = method_parameters
                method = copy(getattr(source_device, method_name))
                # Check if method is meant to be exposed and remove
                # exposed_parameters to prevent double exposing
                if hasattr(method, "exposed_parameters"):
                    del method.__dict__["exposed_parameters"]
                setattr(self, shared_method_name, method)
                self.exposed_methods.append(
                    (shared_method_name, shared_method_attributes)
                )

    def set_session(self, session):
        """
        Set Device's Session.

        :param Session session: Global Session object
        """
        self.session = session
        session.add_device(self)

    def attach_attribute_callback(
        self,
        attribute: Union[List, Tuple, str],
        callback: Callable,
        max_update_rate: Optional[int] = None,
        key_filter: Optional[str] = None,
    ) -> None:
        """
        Attach callback to Attribute.
        By default all changes of every Attribute's key calls callback.
        It's good habit to set key_filter.
        Attach multiple callbacks for different filter.

        If max_update_rate is specified that callback will be called only after
        specified timeout.

        :param attribute: Attribute
        :param callback: callable function or method
        :param max_update_rate: millisecond(s)
        :param key_filter: Callback's filter ie. VALUE, TYPE, ...
        """
        assert callable(callback)
        try:
            attribute_object = self.get_attribute(attribute)
            if attribute_object is None:
                raise Exception(
                    f"couldn't find attribute {attribute}, cannot attach callback"
                )
            attribute_object.attach_callback(callback, max_update_rate, key_filter)

            try:
                self.attribute_attach_map[attribute_object.attribute_id].append(
                    (attribute, callback)
                )
            except KeyError:
                self.attribute_attach_map.update(
                    {attribute_object.attribute_id: [(attribute, callback)]}
                )

            if key_filter is not None:
                callback(attribute_object[key_filter])
            else:
                callback(VALUE, attribute_object[VALUE])
        except (KeyError, AttributeError):
            self.logger.exception("Attribute {} is not defined".format(attribute))

    def detach_attribute_callback(
        self, attribute: Union[List, Tuple, str], callback: Callable
    ) -> None:
        """
        Detach callback.

        :param attribute: Attribute
        :param callback: callable function or method
        """
        assert callable(callback)
        try:
            attribute_object = self.get_attribute(attribute)
            if attribute_object is None:
                raise Exception(
                    f"cannot find attribute {attribute}, cannot detach callback"
                )
            attribute_object.detach_callback(callback)

            try:
                if (attribute, callback) in self.attribute_attach_map[
                    attribute_object.attribute_id
                ]:
                    self.attribute_attach_map[attribute_object.attribute_id].remove(
                        (attribute, callback)
                    )
            except KeyError:
                pass  # Callback does not exists

        except (KeyError, AttributeError):
            if self.logger is not None:
                self.logger.exception("Attribute {} is not defined".format(attribute))

    def _get(self, attribute: Union[str, List, Tuple]) -> Any:
        """
        Private method to obtain attribute from the attributes dict.

        Use this function when You want to get attribute by tuple or list key.
        Example: Device._get((ATTR_STATUS, VALUE))
        Otherwise use faster method Device[ATTR_STATUS][VALUE]

        :param attribute: list
        :return: mixed
        """
        try:
            return get_from_dict(self.attributes, attribute)
        except KeyError:
            return None

    def get_attribute(self, attribute: Union[List, Tuple, str]) -> Any:
        """
        Get Attribute value.

        :param attribute: Attribute
        :return: the attribute if it exists
        """
        try:
            return self._get(Attribute.list_attribute(attribute))
        except KeyError:
            return None

    def get_value(self, attribute: Union[List, Tuple, str], key: str = VALUE) -> Any:
        """
        Get Attribute[key] value.

        :param attribute: Attribute
        :param key: optional key
        :return: mixed
        """
        try:
            return self.get_attribute(attribute)[key]
        except (KeyError, TypeError):
            return None

    def get_raw_value(
        self, attribute: Union[List, Tuple, str], key: str = VALUE
    ) -> Any:
        """
        Get attribute[key] value without offset or factor.

        Use returned value to feed into real hardware.

        :param attribute: Attribute
        :param key: optional key
        :return: mixed
        """
        try:
            attribute_object: Optional[Attribute] = self.get_attribute(attribute)
            if attribute_object is None:
                return None
            return attribute_object.remove_offset_factor(attribute_object[key])
        except (KeyError, TypeError):
            return None

    def _set(
        self,
        attribute: Union[List, Tuple],
        value: Any,
        callback: Optional[Callable] = None,
    ) -> int:
        """
        This sets Device attribute.

        Use this function when You want to set attribute by tuple or list key.
        Example: Device.set((ATTR_STATUS, VALUE), STATUS_IDLE)
        Attribute value is pushed into server if Device is connected listed on one.
        To reduce the amount of pushed attributes we check if value is different from
        the previous one.

        :param attribute: list
        :param value: Value of attribute
        :return: Set token
        """
        attribute_object: Attribute = get_from_dict(self.attributes, attribute[:-1])
        if not attribute_object.allow_broadcast():
            allow_broadcast = False
        elif attribute[-1] == VALUE and attribute_object.with_set_function:
            allow_broadcast = attribute_object.set_value_when_set_function
        else:
            allow_broadcast = True
        if allow_broadcast:
            attribute_object.last_broadcast_time = time.perf_counter_ns()

        current_value = attribute_object[attribute[-1]]
        attribute_object[attribute[-1]] = value

        set_token = attribute_object.read_and_reset_token()
        if not set_token and callable(callback):
            callback(attribute, value)

        if allow_broadcast and kamzik3.session is not None:
            value = attribute_object[attribute[-1]]
            if not is_equal(current_value, value):
                if self.session and attribute_object[SAVE_CHANGE]:
                    self.session.update_config(self.device_id, attribute, value)
                    self.config["attributes"][tuple(attribute)] = value
                if kamzik3.session.publisher is not None:
                    header = "{}.{}".format(
                        attribute_object.attribute_id, attribute[-1]
                    )
                    kamzik3.session.publisher.push_message(header, value)

        return set_token

    def set_attribute(
        self,
        attribute: Union[List, Tuple, str],
        value: Any,
        callback: Optional[Callable] = None,
    ) -> int:
        """
        This sets Device attribute.

        Re-implement it for client or any special use.

        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :return: token
        """
        list_attribute = Attribute.list_attribute(attribute)
        if list_attribute[-1] == VALUE:
            return self.set_value(list_attribute[:-1], value, callback)
        else:
            return self._set(list_attribute, value, callback)

    def set_value(
        self,
        attribute: Union[List, Tuple, str],
        value: Any,
        callback: Optional[Callable] = None,
        key: Any = VALUE,
    ) -> int:
        """
        Set attribute value.

        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :param key: Attribute's key
        :return: token
        """
        attribute_list = Attribute.list_attribute(attribute)
        attribute_object = self.get_attribute(attribute_list)
        with SetFunction(attribute_object, callback):
            return self._set(attribute_list + [key], value, callback)

    def set_offsetted_value(
        self,
        attribute: Union[List, Tuple, str],
        value: Any,
        callback: Optional[Callable] = None,
        key: Any = VALUE,
    ) -> None:
        """
        Apply the offset and factor to the attribute value obtained from Device, before
        the value is stored in the key VALUE.

        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :param key: Attribute's key
        """
        list_attribute = Attribute.list_attribute(attribute)
        attribute_object = self.get_attribute(list_attribute)
        if attribute_object is None:
            raise Exception(
                f"couldn't get attribute {attribute}, couldn't set offsetted value"
            )
        value = attribute_object.apply_offset_factor(value)
        self._set(list_attribute + [key], value, callback)

    def set_raw_value(
        self,
        attribute: Union[List, Tuple, str],
        value: Any,
        callback: Optional[Callable] = None,
        key: Any = VALUE,
    ) -> None:
        """
        Set attribute value obtained from Device.
        You want to call this to set Value obtained directly from Device.
        Apply offset and factor before value is stored in VALUE key.

        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :param key: Attribute's key
        """
        warnings.warn(
            "set_raw_value is deprecated, use set_offsetted_value instead",
            DeprecationWarning,
        )
        list_attribute = Attribute.list_attribute(attribute)
        attribute_object = self.get_attribute(list_attribute)
        if attribute_object is None:
            raise Exception(
                f"couldn't get attribute {list_attribute}, couldn't set raw value"
            )
        value = attribute_object.apply_offset_factor(value)
        self._set(list_attribute + [key], value, callback)

    # pylint: disable=unused-argument
    def connect(self, *args):
        """
        Connect to Device via interface.
        This method should be re-implemented when using ETH or RS232, ...

        :param mixed args: connect attributes
        """
        try:
            self.connecting = True
            self.connected = False
            self.device_connection_poller.add_connecting_device(self)
            self.handle_connect_event()
        except DeviceError:
            self.logger.exception("Connection exception")

    def handle_readout_callback(
        self,
        callback: Optional[Callable],
        attribute: Union[List, Tuple, str],
        output: Any,
    ) -> None:
        """
        This method handle readout from Device.
        It's called when Device replies to Request.
        This method should be re-implemented when using ETH or RS232, ...

        :param callback: callable
        :param attribute: tuple, list, str
        :param output: Reply from Device
        """
        if callback is None:
            return
        else:
            assert callable(callback)
            callback(attribute, output)

    def handle_connect_event(self):
        """
        Handle connect event.
        It's called when Device is successfully connected.
        This method should be re-implemented when using ETH or RS232, ...
        """
        try:
            self.handle_connect()
            self.connected = True
            self.connecting = False
            self.handle_configuration_event()
        except DeviceError:
            self.logger.exception("Error during connection")

    def handle_connect(self):
        """
        Handle connection.
        Measure and write connection time and set request and response timestamps.
        When re-implementing, call this method at the beginning.
        """
        self.set_status(STATUS_CONNECTED)
        self.logger.info(
            "Device connection took {} sec.".format(time.time() - self.init_time)
        )
        self.request_timestamp = self.response_timestamp = time.time()

    def handle_configuration_event(self):
        """
        Handle configuration event.
        Device is connected, now we need to configure it.
        """
        try:
            self.set_status(STATUS_CONFIGURING)
            self.handle_configuration()
        except DeviceError:
            self.logger.exception("Error during configuration")

    def handle_configuration(self):
        """
        This is placeholder method for handle configuration.
        This method must be re-implemented as every Device has special configuration.
        Example of implementation:

            start_at = time.time()

            def _finish_configuration(*_, **__):
                self._config_commands()
                self._config_attributes()
                self.start_polling()
                self.set_status(STATUS_CONFIGURED)
                self.logger.info(
                    u"Device configuration took {} sec.".format(time.time() - start_at)
                )

            _finish_configuration()
        """
        raise NotImplementedError("Must be implemented in subclass")

    def handle_connection_error(self, message=None):
        """
        Method to handle Device's connection error.

        :param str message: cause of connection error
        """
        self.logger.error(message)
        self.connection_error = True
        self.close()

    def handle_response_error(self, message=None):
        """
        Method to handle Device's response error.

        :param str message: cause of response error
        """
        self.logger.error(message)
        self.response_error = True
        self.close_connection()

    def handle_command_error(self, readout_command, readout_output):
        """
        Method to handle Device's command error.

        :param str readout_command: original command
        :param str readout_output: error response from device
        """
        self.set_value(ATTR_LAST_ERROR, str(readout_output))
        self.logger.error(
            "Command error\nCommand: {!r}\nOutput: {!r}\nCommand buffer: {!r}".format(
                readout_command, readout_output, self.commands_buffer
            )
        )

    def handle_observer_attached(self, observer):
        """
        Callback on Observer Detached event.

        :param Observer observer: Observer
        """
        observer.subject_update(ATTR_STATUS, self.get_value(ATTR_STATUS), self)

    def handle_observer_detached(self, observer):
        """
        Callback on observer Attached event.

        :param Device observer: Observer
        """

    def handle_readout(self, readout_buffer) -> tuple:
        """
        We have data in readout_buffer.
        First pop from commands_buffer to get metadata of the command.
        Calculate command latency and push it to the buffer.

        :param str readout_buffer: Response to request from Device
        :return str: Attribute, Response, Callback, Token
        """
        self.response_timestamp = time.time()

        try:
            (
                attribute,
                token,
                callback,
                _returning,
            ), command_init_timestamp = self.commands_buffer.popleft()
            latency = (time.time() - command_init_timestamp) * 1000
            self.latency_buffer.append(latency)
            if len(self.latency_buffer) == self.latency_buffer.maxlen:
                self.set_attribute(
                    (ATTR_LATENCY, VALUE),
                    sum(self.latency_buffer) / self.latency_buffer.maxlen,
                )
                self.set_value(ATTR_HANGING_COMMANDS, len(self.commands_buffer))
                self.latency_buffer.clear()
        except IndexError:
            self.handle_response_error(
                "Trying to pop from empty command buffer. "
                "Content of readout buffer is: {}".format(readout_buffer)
            )
            return RESPONSE_ERROR, "", None, False

        if self.command_encoding is None:
            return attribute, readout_buffer, callback, token
        else:
            return attribute, "".join(readout_buffer), callback, token

    def disconnect(self) -> bool:
        """
        Call this function to cleanly close connection.

        :return bool: True if successful
        """
        if self.closing:
            return False
        elif self.connected:
            self.close_connection()
        elif self.connecting:
            self.close()
        return True

    def close_connection(self):
        """
        Call this function when closing connected Device.

        Stop polling and continue with close() method.
        """
        self.stop_polling()
        self.close()

    def close(self) -> bool:
        """
        When Device is close(), set all connection flags to False.
        Device is now Disconnected from Physical interface.
        Close socket, close port, etc...
        :return bool: True if successful
        """
        if self.closing:
            return False
        else:
            self.closing = True
            self.connected = False
            self.set_status(STATUS_DISCONNECTED)
            self.closing = False
            self.connected = False
            return True

    def reconnect(self) -> bool:
        """
        Reconnect devices.
        :return bool: True if successful
        """
        mapping = self.yaml_mapping()
        self.__init__(**mapping)  # type: ignore
        # rather than calling reconnect and re-initializing the class, it would be
        # better to re-instantiate the device
        return True

    def set_status(self, status):
        """
        Set attributes status value.
        There is no restriction for STATUS value.
        But to ensure Device is fully compatible it should be one of the following:
            STATUS_DISCONNECTED: Device is not connected (not ready)
            STATUS_CONNECTING: Device called connect() method (not ready)
            STATUS_CONNECTED: Device is connected (not ready)
            STATUS_CONFIGURING: Device is configuring (not ready)
            STATUS_CONFIGURED: Device is configured (ready)
            STATUS_IDLE: Device is IDLE (finished previous task - ready)
            STATUS_BUSY: Device is performing a task (not ready)
            STATUS_ERROR: Device is in ERROR state (not ready)
            STATUS_DISCONNECTING: Device is disconnecting (not ready)
        STATUS_IDLE and STATUS_CONFIGURED can be interchanged.
        But STATUS_CONFIGURED server as a metadata indicating, that it was configured at
        least once.

        :param str status: Status of Device
        """
        self.set_value(ATTR_STATUS, status)

    def is_status(self, status) -> bool:
        """
        Compare status and current Device status

        :param str status: Status to compare
        :return bool: True if equal
        """
        return bool(self.get_value(ATTR_STATUS) == status)

    def in_statuses(self, statuses: Union[List, Tuple]) -> bool:
        """
        Check list of statuses against current device status.

        :param statuses: List of statuses
        :return: True if current status is in statuses
        """
        return self.get_value(ATTR_STATUS) in statuses

    def start_polling(self):
        """
        This method should be re-implemented.
        Add all polled commands here.
        Call this method at the beginning to reset response and request timestamp.

        Example:
            self.devicePoller.add(devices, command, time in milliseconds)
            self.devicePoller.add(self, 'cmd0', 400)
        """
        self.response_timestamp = self.request_timestamp = time.time()

    def stop_polling(self):
        """Remove Device from device_poller loop."""
        self.device_poller.stop_polling(self)

    def push(self, data):
        """
        Send command to Device.
        This method should be reimplemented for specific interface.
        Use command() for sending commands from Client.

        :param str data: Request to Device
        """

    def command(
        self,
        command: str,
        callback: Optional[Callable] = None,
        with_token: bool = False,
        returning: bool = True,
    ) -> Optional[int]:
        """
        Check command for validity and push it to the poller loop.

        !!! Use only this function to send command to Device. !!!
        Commands are stored into buffer and are flushed on poller tick.
        Command will be executed on next poller tick.
        Attach token if with_token is True.
        Handle request ONLY when returning is True.
        !!! There are Devices, that silently executes command and in that case we are
        not waiting for response. !!!

        :param command: Request to Device
        :param callback: callback after Device replies to Command
        :param with_token: token
        :param returning: bool
        :return: token
        """
        if not self.valid_command_format(command):
            raise CommandFormatException("Command '{}' form is invalid".format(command))

        if self.connected or self.connecting:
            token = 0
            if with_token is True:
                token = next(self.token_generator)
            elif with_token >= 1:
                token = with_token
            self.device_poller.prepare_command(
                self, (command, token, callback, returning)
            )
            return token

        return None

    def remove(self):
        """Remove Device completely from Server and Session."""
        self.disconnect()

        if self.logger is not None:
            for handler in self.logger.handlers[:]:
                handler.close()

        self.logger = None
        if self.session is not None:
            self.session.remove_device(self)

        for cb_tuples in self.attribute_attach_map.values():
            for cb_tuple in cb_tuples:
                self.detach_attribute_callback(*cb_tuple)

        self.attribute_attach_map = None
        self.attributes = None
        self.exposed_methods = None
        self.token_generator = None
        self.connecting_time = None
        self.response_timestamp = None
        self.request_timestamp = None
        self.connected = None
        self.connecting = None
        self.closing = None
        self.closed = None
        self.connection_error = None
        self.response_error = None
        self.commands_buffer = None

    def valid_command_format(self, command) -> bool:
        """
        Check if command format is valid.

        Return False or optionally raise CommandFormatException.
        Re-implement this method for any other device.

        :param str command: Input command
        :return bool: True if has valid format
        """
        return command is not None

    def send_command(self, commands: List[Any]) -> List[Any]:
        """
        Send command using defined interface (TCP/IP, RS232, ...).
        !!! Make sure to call it directly when it's needed, for example for Connection
        handshake.!!! Unicode command is always .encode() into ASCII before it's send to
        the interface.

        :param list commands: list of commands to be send
        :return list: list of NOT sent commands
        """
        if not self.accepting_commands():
            return commands
        try:
            mark = time.time()
            if self.connected:
                command_data = b"" if self.command_encoding is None else ""
                commands_to_push_counter = 0
                while commands:
                    command = commands.pop(0)
                    commands_to_push_counter += 1
                    command_data += command[0]
                    if len(command_data) > self.push_buffer_size:
                        commands.insert(0, command)
                        command_data = command_data[: -len(command[0])]
                        break
                    if command[3]:  # If command is returning
                        self.commands_buffer.append((command, mark))
                    else:  # Command is not returning, simulate immediate execution
                        self.response_timestamp = mark
                    if commands_to_push_counter >= self.push_commands_max:
                        break
                if self.command_encoding is None:
                    # command_data is of type bytes
                    self.push(command_data)
                else:
                    if not isinstance(command_data, str) or not isinstance(
                        self.command_encoding, str
                    ):
                        raise TypeError(
                            "command_data and command_encoding should be strings"
                        )
                    self.push(command_data.encode(self.command_encoding))
            else:
                self.handle_response_error("Device is not connected")
        except IndexError:
            self.handle_connection_error(
                "Device {} buffer error".format(self.device_id)
            )
        except (WriteException, serial.SerialException):
            self.handle_response_error("Device {} writing error".format(self.device_id))
        finally:
            # pylint: disable=lost-exception
            return commands

    def accepting_commands(self) -> bool:
        """
        Check if device is ready to send another commands.

        :return: bool
        """
        self.request_timestamp = time.time()
        if not self.connected or len(self.commands_buffer) > 0:
            return False
        else:
            return True

    def poll_command(self, command, interval):
        """
        Add command to device poller loop to be polled in desired interval.

        :param str command: Request / Command
        :param int interval: interval in ms
        """
        self.device_poller.add(self, command, interval)

    def remove_poll_command(self, command, interval):
        """
        Remove command from device poller loop.

        :param str command: Command to remove
        :param interval: interval in ms
        """
        self.device_poller.remove(self, command, interval)

    def _config_attributes(self):
        """
        Set attributes to desired values found in config.
        Always call this from handle_configuration method.
        """
        if self.config:
            for attribute, value in self.config.get("attributes", {}).items():
                unit = self.get_attribute(attribute[:-1])[UNIT]
                if unit in (None, "") or attribute[-1] == UNIT:
                    self.set_attribute(attribute, value)
                else:
                    if isinstance(value, list):
                        set_value = []
                        for v in value:
                            try:
                                set_value.append(
                                    device_units(self, list(attribute[:-1]), v).m
                                )
                            except (UndefinedUnitError, DimensionalityError):
                                set_value.append(v)
                    else:
                        try:
                            set_value = device_units(
                                self, list(attribute[:-1]), value
                            ).m
                        except (UndefinedUnitError, DimensionalityError):
                            set_value = value
                    self.set_attribute(attribute, set_value)

    def _config_commands(self):
        """
        Execute required methods found in commands section of config
        Always call this from handle_configuration method.
        """
        if self.config:
            for command in self.config.get("commands", []):
                self.command(command)

    def _expose_methods_to_clients(self):
        """Expose all decorated methods with @expose_method to the Client."""
        for method in dir(self):
            method = getattr(self, method)
            if (
                callable(method)
                and hasattr(method, "exposed_parameters")
                and method.__name__ not in self.exposed_methods
            ):
                self.exposed_methods.append(
                    (method.__name__, method.exposed_parameters)
                )

    def wait_for_status(
        self,
        statuses: List,
        retry_timeout: int = 1000,
        throw_exception: bool = True,
        callback: Optional[Callable] = None,
    ) -> Union[bool, Thread]:
        """
        Wait until device change it's status in one defined in statuses.
        Method is non blocking, when callback is set.

        :param statuses: List of one or multiple statuses to check
        :param retry_timeout: Maximum time in ms to wait for status in ms
        :param throw_exception: Raise exception when True
        :param callback: Function to call after check is finished
        :return: True if device is in one of statuses
        """

        def _wait(_callback=None):
            success = self.in_statuses(statuses)
            retry_counter = 0
            while not success:
                if not self.connected:
                    break
                time.sleep(0.05)
                retry_counter += 50
                success = self.in_statuses(statuses)
                if retry_counter >= retry_timeout:
                    break
            if _callback is not None:
                _callback(success)
            elif throw_exception:
                if not success:
                    raise DeviceError(
                        "Status was not set to {} within timeout of {} ms".format(
                            statuses, retry_timeout
                        )
                    )
                else:
                    return True
            else:
                return success

        if callback is None:
            return bool(_wait())
        else:
            thread = Thread(target=_wait, args=[callback])
            thread.start()
            return thread

    def wait_until_value_set(
        self,
        attribute: Union[List, Tuple, str],
        value: Any,
        retry_timeout: int = 1000,
        throw_exception: bool = True,
        callback: Optional[Callable] = None,
    ) -> Union[bool, Thread]:
        """
        Trigger the callback or block until the attribute is set to the requested value.

        There is retry_timeout of 1000ms until DeviceError is raised.

        :param attribute: Attribute to check
        :param value: Value to check
        :param retry_timeout: Retry check timeout in ms
        :param throw_exception: Raise exception when True
        :param callback: Function to call after check is finished
        :return: True if Attribute's value is equal
        """

        def _wait(_callback=None):
            attribute_object = self.get_attribute(attribute)
            success = attribute_object[VALUE] == value
            retry_counter = 0
            while not success:
                if not self.connected:
                    break
                time.sleep(0.05)
                retry_counter += 50
                success = attribute_object[VALUE] == value
                if retry_counter >= retry_timeout:
                    break
            if _callback is not None:
                _callback(success)
            elif throw_exception:
                if not success:
                    raise DeviceError(
                        "Value of {} was not set to {} within timeout of {} ms".format(
                            attribute, value, retry_timeout
                        )
                    )
                else:
                    return True
            else:
                return success

        if callback is None:
            return bool(_wait())
        else:
            thread = Thread(target=_wait, args=[callback])
            thread.start()
            return thread

    def is_alive(self) -> bool:
        """
        Check if device is still connected.

        :return bool: True if Device is connected
        """
        if not self.connected or self.closing or self.response_error:
            return False
        time_diff = self.request_timestamp - self.response_timestamp
        if time_diff >= self.response_timeout * 1e-3:
            self.handle_response_error("Response timeout")
            return False
        return True

    @expose_method()
    def server_disconnect(self):
        """This function serves as a disconnect wrapper for Client"""
        self.disconnect()

    @expose_method()
    def server_reconnect(self):
        """This function serves as a reconnect wrapper for Client"""
        self.reconnect()

    @expose_method()
    def get_config(self) -> Any:
        """This function return Device configuration."""
        return yaml.dump(self.config, Dumper=yaml.Dumper)
